# Use this when you want to extend the core with third party modules and you need an api
# package to include to use all of the package functions.

#
# This represents the underlying language architecture with which we build the application
# package out of.  In otherwords, all of the basic, system and utilitarian functions and
# objects are brought in by this module to unify their particular platform-specific imple-
# mentations with the object definitions of the package model.

# It probably makes sense to import these things specifically into the different features
# of the package, but we're trading off detailed provisioning with a more broad platform
# compatibilty design.  The only thing a detail provisioning does for us at this point is
# decouple the features from the application instead of preloading everything at boot.
#
# Also, to pre-import might mean committing to a certain implementation in terms of what
# components this module represents for the larger application product.

# Just import all of the platform objects here, and then import what you need in modules.

from stuphos import getConfig, getConfigObject, getSection, enqueueHeartbeatTask
from stuphos import logException, log as mudlog, inline

from stuphos import db
from stuphos.db import dbCore

from stuphos.runtime import Synthetic, Object, Undefined, LookupObject
from stuphos.runtime.facilities import Facility
from stuphos.runtime.registry import getObject

from stuphos.runtime.architecture import Procedure, Computer, Source, Yield, Continuation
from stuphos.runtime.architecture import OuterFrame, BypassReturn, Generator, YieldFrame
from stuphos.runtime.architecture import NoAccessException, InterpreterState, Uncompleted
from stuphos.runtime.architecture import newComponent, writable, writeprotected, blockSetAttr
from stuphos.runtime.architecture import wrapper, representable, Transparent, Translucent
from stuphos.runtime.architecture import ExceptionType, baseExceptionValue
from stuphos.runtime.architecture import extension, extensionObject, baseExceptionValue
from stuphos.runtime.architecture import baseInstance, safeNativeClass, _safe_native, _safe
from stuphos.runtime.architecture import safeNativeClass as _safe_native_class
from stuphos.runtime.architecture import safeNativeObject, _safe_native_object
from stuphos.runtime.architecture import _safe_iterator, ScheduledProcedureMixin
from stuphos.runtime.architecture import safeNativeNoKeyword, isSpecialPrincipal

from stuphos.runtime.architecture.routines import apply

from stuphos.etc import isYesValue, isNoValue, nth
from stuphos.etc import reraise, reraiseSystemException
from stuphos.etc import getSystemException, getSystemExceptionString

from stuphos.etc.tools import ordereddict, getSystemException, isYesValue
from stuphos.etc.tools.timing import date as now, Elapsed
from stuphos.etc.tools.misc import pickleSafe

from stuphos.db.orm import (GirlNode, GirlBootOrderNode, GirlPermission,
							GirlPrincipalGroup, PrincipalRelation)

from stuphmud import server as mud
from phsite.network import renderTemplate


# Dynamically-linked symbols.  Because of complex imports, and now the separation of
# these packages, some symbols need to be loaded dynamically.  This means a certain
# re-architecting of depending code in some cases where it is previously done at the
# beginning of its module initialization.

def getWebProgrammer():
	from phsite.network.embedded.olc import WebProgrammer
	return WebProgrammer

def getIsSuperuser():
	from phsite.network import is_superuser
	return is_superuser

def getNoAccessTemplate():
	from phsite.network.embedded.olc.base import NOACCESS_TEMPLATE
	return NOACCESS_TEMPLATE

def getMobileWrapper():
	from phsite.network.embedded.olc.triggers import MobileWrapper
	return MobileWrapper

def getFindClaimedPlayers():
	from phsite.network.embedded.olc import findClaimedPlayers
	return findClaimedPlayers

def getOn():
	from stuphmud.server import on
	return on # mud.on

def getPlayerModel():
	from phsite.network.models.realm import Player
	return Player

def getDefaultPlayerModel():
	from phsite.network.models.realm import DefaultPlayer
	return DefaultPlayer

def getUsageModel():
	from phsite.network.models.realm import Usage
	return Usage

def getFindUserByName():
	from phsite.network.accounts import findUserByName
	return findUserByName

def getFindUser():
	from phsite.network.accounts import findUser
	return findUser

def getFindCurrentUser():
	from phsite.network.accounts import findCurrentUser
	return findCurrentUser


# Some particular things cannot be imported at boot time, like the task processor
# base class that is actually depending on some derivative functionality.. which
# is wrong, but I will let that feature's development determine how to resolve
# itself because for now, the only solution is to just leave it unchanged and out
# of this architectural module.

# from ph.language.interpreter import native, objects as girlObjects, Girl, Script
# from ph.language.interpreter.objects import GirlSystemModule


# These things can't exactly be imported into an architecture pretty much because of
# dependencies, which primarily means that those functions must specifically import
# their components from other parts of THIS package.  As you can see, they're all
# relative, so I pretty leave them unchanged.

# from . import Object, Procedure, Heartbeat, Scheduling, tasklog, billing
# from .heartbeat import Heartbeat

# from .vtpool import VTPool as RealtimeAPI
# from ..virtual import Machine as VM, Native, Pure


# Some modules in this package shouldn't actually be here.  They rely on these things:

# from ph.network.server.player.interfaces.dictcmd import CmdDict

# from ph.emulation import startTask




# This should go in an implementor package, actually.
from xmlrpc.client import ServerProxy, Transport, Fault

def getXmlRpcPythonEvaluator(*args, **kwd):
    return XmlRpcPythonShell(*args, **kwd).evaluateStatement

class AuthTransport(Transport):
    ##    def __init__(self, cookiejar, use_datetime = 0):
    ##        Transport.__init__(self, use_datetime = use_datetime)
    ##        self.cookiejar = cookiejar

    # Use installed cookiejar with urllibl when operating the request.
    ##    def single_request(self, host, handler, request_body, verbose=0):
    ##        pass

    pass

class XmlRpcPythonShell:
    PROTOCOL = 'http'
    PATH = 'RPC2'

    def __init__(self, hostname, port, path = None):
        if path is None:
            path = self.PATH

        transport = None # AuthTransport(cookiejar)
        self.server = ServerProxy('%s://%s:%s/%s' % (self.PROTOCOL,
                                                     hostname, port,
                                                     path),
                                  allow_none = True,
                                  transport = transport)

    def evaluateStatement(self, shell, peer, argstr):
        # todo: a way to exit this evaluator...
        # this should go in the managing command..?
        # xmlrpc-python-connect
        if argstr == 'exit!':
            peer.interpreter.popEvaluator()
            return True

        # todo: asynchronously!!
        try: result = self.server.evaluate(argstr)
        except Fault as fault:
            peer.page_string(fault.faultString)
        else:
            if result:
                if not isinstance(result, str):
                    result = str(result)

                peer.write(result)

        return True

try: from stuphmud.server.player import ACMD
except ImportError:
    ##    from mud.tools import HandleCommandError
    ##    HandleCommandError(full_traceback = True)

    ##    from traceback import print_exc as traceback
    ##    traceback()

    pass

else:
    COMMAND_NAME = 'xmlrpc-python-connect'

    @ACMD(COMMAND_NAME + '*')
    def doXmlRpcPythonConnect(peer, cmd, argstr):
        args = argstr.split() if argstr else ()
        path = None

        if len(args) is 2:
            (hostname, port) = args
        elif len(args) is 3:
            (hostname, port, path) = args
        else:
            print('Usage: %s <hostname> <port> <path>' % COMMAND_NAME, file=peer)
            return True

        pushE = getattr(peer.interpreter, 'pushEvaluator', None)
        if callable(pushE):
            pushE(getXmlRpcPythonEvaluator(hostname, port, path))
            print('Opening XmlRpc Context for [%s:%s]%s%s' % (hostname, port,
                                                                         path and ' on ' or '',
                                                                         path or ''), file=peer)
        else:
            print('Your shell interpreter does not support new evaluators.', file=peer)

        return True



# Application component library.
# Copyright 2010-2021 runphase.com .  All rights reserved.
# --
# The core library made out of components and frameworks for building applications.
# Applications are rather specific, but most of the system or platform features are
# provided for as component layers.  And although specific server architectures
# might be more heavily rooted in peripheral projects, much of that framework is
# provided by the contained network components.  Most everything works towards the
# purposes of the application.

# Relative imports are done for modules within the same subdirectory under any
# single one of the top-level stuphos subpackages.  Between code in these top-
# level packages, an absolute import should be used.

# This should be removed to the specific implementation configuration compartment.
from .runtime.core import *


# try: wmc = io.here.resources('site.wmc').loading.structure
# except: pass
# else: globals().update(dictOf(wmc))



# Player Interpreter.
from stuphmud.server.player import HandleCommandError
from .interfaces import *

from stuphos.triggers import verbCommand

# Todo: implement a prompt-response-handler queue.
class ShellI(CommandManager, HistoryManager, CodeManager, ThreadingManager):
    "Hot Command Interpreter [exit! to exit, ;%`<python statement>, ;%`'[<edit program name>]]"

    def __init__(self, commands, scope):
        CommandManager.__init__(self, commands)
        CodeManager.__init__(self, *makeCodeEvaluators(scope))

    def shellCommand(self, peer, line):
        # Fraun Dec 21st 2005 - New form
        cmd, argstr = self.parseCommand(line)
        if not cmd:
            return False

        # Fraun Sept 21st 2006 - History override.
        cmd, argstr = self.historyOverride(peer, cmd, argstr)
        if cmd is None:
            return True

        # Fraun Dec 21st 2005 - Incorporation of CmdDict
        if self.playerCommand(peer, cmd, argstr):
            return True

        # Parse regular command-names
        # XXX Why would this be allowed at this point...?
        ##    if cmd.lower() == 'exit!':
        ##        del peer.interpreter
        ##        return True

        # Fraun Nov 4th 2005 - Only do python if preceeded by ';'.  Do not trigger chat.
        if cmd == ';' or cmd == '`' or cmd == '%':
            # Fraun Sep 20th 2005 - Delegate
            return self.evaluateStatement(peer, argstr)

        # Fraun Feb 20th 2018 - MOO-like verb commands.  This should go in CommandManager.
        # Fraun Nov 14th 2020 - Moved to after all other command handlers because it shall
        # technically fire off a girl procedure which will handle unknown commands.
        if verbCommand(peer, cmd, argstr):
            return True

        return False


    # These are defined for convenience, but unused (Feb 27th 2022).
    @classmethod
    def evaluateCommand(self, *args, **kwd):
        from stuphmud.server.player.interfaces.code.girl import PlayerScript
        return PlayerScript.evaluateCommand(*args, **kwd)

    @classmethod
    def evaluateVerb(self, *args, **kwd):
        from stuphmud.server.player.interfaces.code.girl import PlayerScript
        return PlayerScript.evaluateVerb(*args, **kwd)


    def withPeerHeadAndException(self, peer, exception, procedure, *args, **kwd):
        '''
        Evaluate procedure while output is going to peer.  Send exceptions to peer.
        Not all exceptions, let those of kind <exception> through.
        '''
        import sys

        # Todo: better management of system state.
        STDOUT = sys.stdout
        STDERR = sys.stderr

        setattr(sys, '__peer_shell_stdout', STDOUT)
        setattr(sys, '__peer_shell_stderr', STDERR)

        # stdout queues to peer output buffer
        sys.stdout = peer
        sys.stderr = peer

        # What about stdin?

        try:
            result = procedure(peer, *args, **kwd)
            sys.stdout = STDOUT
            sys.stderr = STDERR
            return result

        except:
            sys.stdout = STDOUT
            sys.stderr = STDERR

            (etype, value, tb) = sys.exc_info()
            if exception is not None and isinstance(value, exception):
                raise etype

            return HandleCommandError(peer, (etype, value, tb))

    def withPeerHead(self, peer, procedure, *args, **kwd):
        return self.withPeerHeadAndException(peer, None, procedure, *args, **kwd)

    def __call__(self, peer, line):
        'Interpret player input line by executing it in this shell.'

        # If someone points their web browser at this port, redirect them.
        from stuphmud.server.player import http
        if http.detectHttpRequest(peer, line):
            return

        # Synthesize Event
        if playerInput(peer, line):
            return

        # Invoke "CGI"
        return self.withPeerHead(peer, self.shellCommand, line)

        # debugOn()
        # return self.shellCommand(peer, line)

        # try:
        #   with Head(peer):
        #       self.shellCommand(peer, line)
        # except:
        #   mud.player.HandleCommandError(peer, sys.exc_info())

    def __repr__(self):
        return "<Python Player Shell Interpreter>"

from time import time as now
def getDataRate(p):
    duration = float(now() - p.login_time)
    return (p.bandwidth_in / duration,
            p.bandwidth_out / duration)

from stuphos.runtime import eventResult
def makePrompt(peer):
    # This should go in mud.player again.
    if peer.state == 'Playing':
        makePrompt = getattr(peer, 'makePrompt', None)
        if callable(makePrompt):
            return eventResult(makePrompt())

        p = peer.prompt # Why would this return None?
        a = peer.avatar
        if a is not None:
            if a.hasMail:
                p = '(&wMail&N) %s' % p

            if False: # a.properties.display.prompt.flags.Bandwidth:
                p = '(&M%.1f&N:&m%.1f&N/s) %s' % (getDataRate(peer) + (p,))

        return eventResult(p)

class PromptFor:
    def __init__(self, peer, callback, message = None, write_mode = False,
                 compact_mode = True, timeout = None):

        # Todo: allow for interacting with ShellI and installing a sub-interpreter.
        # This allows the main player-shell to operate some of the lower-level features
        # like interpreted code while calling this prompting mechanism when needed.
        # (good for debugging the prompt)

        self.previous_interpreter = peer.interpreter
        self.callback = callback

        a = peer.avatar
        if a is not None:
            if write_mode:
                pf = a.playerflags
                self.previous_write_mode = pf.Writing
                pf.Writing = True
            else:
                self.previous_write_mode = None

            if compact_mode:
                pf = a.preferences
                self.previous_compact_mode = pf.Compact
                pf.Compact = True
            else:
                self.previous_compact_mode = None

        peer.interpreter = self

        self.message = message
        if message is not None:
            self.previous_make_prompt = getattr(peer, 'makePrompt', None)
            peer.makePrompt = (lambda:message)

    def regenerate(self, peer):
        self.__class__(peer, self.callback, message = self.message,
                       write_mode = self.previous_write_mode is not None)

    def __call__(self, peer, line):
        if self.previous_interpreter is None:
            del peer.interpreter
        else:
            peer.interpreter = self.previous_interpreter

        if self.previous_make_prompt is None:
            del peer.makePrompt
        else:
            peer.makePrompt = self.previous_make_prompt

        a = peer.avatar
        if a is not None:
            pwm = self.previous_write_mode
            if pwm is not None:
                a.playerflags.Writing = pwm

            pcm = self.previous_compact_mode
            if pcm is not None:
                a.preferences.Compact = pcm

        # Q: Catch exceptions?  Failure should still consume the action.
        # XXX Breaking backwards compatability: pass in this instance to the
        # callback (natural solution, keeps things current), otherwise, the
        # new PromptFor instance is completely separated from the controller.
        if self.callback(self, peer, line):
            self.regenerate(peer)

        return True

    def __repr__(self):
        return '<%s callback: %r, previous: %r>' % (self.__class__.__name__,
                                                    self.callback,
                                                    self.previous_interpreter)

def Prompting(peer, message, *args, **kwd):
    def makePromptResponse(function):
        PromptFor(peer, message = message, callback = function)
        return function

    return makePromptResponse


from stuphos.runtime import DeclareEvent
from stuphos import getBridgeModule

class playerInput(DeclareEvent):
    Module = getBridgeModule()



'''
[Services]
facility.lo-idle: stuphos.management.idling.IdleConnection

[Network:Idling]
port: 4000
name: idle
password: none
login-delay: 3
login-sequence:
	ok

reconnect: false

'''

from stuphos.runtime.facilities import Facility
from stuphos.etc.tools import isYesValue
from stuphos import getSection

from telnetlib import Telnet
from threading import Thread
from time import sleep

def stripNL(line):
	while line.endswith('\n'):
		line = line[:-1]

	return line

class IdleConnection(Facility):
	'Outgoing loopback connection.'

	NAME = 'Network::Idle'

	class DEFAULTS:
		PORT = 4000
		RECONNECT = False # True = Aggressively.
		DELAY = 2

	@classmethod
	def create(self):
		return self(getSection('Network:Idling'))

	def __init__(self, cfg):
		self.port = cfg.get('port', self.DEFAULTS.PORT)
		self.name = cfg['name']
		self.password = self.parse_loadAuth(cfg)
		self.login_sequence = self.parse_loginSequence(cfg.get('login-sequence', ''))
		self.reconnectCfg = isYesValue(cfg.get('reconnect', self.DEFAULTS.RECONNECT))
		self.login_delay = int(cfg.get('login-delay', self.DEFAULTS.DELAY))

		assert self.name

		self.start()

	def parse_loadAuth(self, cfg):
		name = cfg['auth-file']
		if name is None:
			auth = cfg['password']
			return auth

		return stripNL(open(name).read())

	def parse_loginSequence(self, sq):
		return sq.split('\n')

	def reconnect(self):
		return self.reconnectCfg

	def loginSequence(self):
		yield self.name
		yield self.password

		yield from self.login_sequence

	def login(self):
		self.telnet = Telnet('127.0.0.1', self.port)

		for message in self.loginSequence():
			self.sendln(message)

	def sendln(self, message):
		message = message.encode()
		self.telnet.write(message + b'\r\n')


	def start(self):
		Thread(target = self.run).start()


	def run(self):
		while True:
			sleep(self.login_delay)

			try:
				self.login()
				self.sessionCycle()

			except Exception as e:
				print(f'[{self.NAME}] {e.__class__.__name__}: {e}')

				if self.reconnect():
					continue

			break


	def sessionCycle(self):
		while True:
			self.handleInput(self.read())

	def read(self):
		# Goal: prevent thread spin.
		return self.telnet.read_some() # Read at least one byte or EOF; may block.

	def handleInput(self, input):
		pass




# Copyright Google.
# Ammalgamation to define OrderedDict, namedtuple

# Copyright 2007 Google, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""Abstract Base Classes (ABCs) according to PEP 3119."""


# Instance of old-style class
class _C: pass
_InstanceType = type(_C())


def abstractmethod(funcobj):
    """A decorator indicating abstract methods.

    Requires that the metaclass is ABCMeta or derived from it.  A
    class that has a metaclass derived from ABCMeta cannot be
    instantiated unless all of its abstract methods are overridden.
    The abstract methods can be called using any of the normal
    'super' call mechanisms.

    Usage:

        class C:
            __metaclass__ = ABCMeta
            @abstractmethod
            def my_abstract_method(self, ...):
                ...
    """
    funcobj.__isabstractmethod__ = True
    return funcobj


class abstractproperty(property):
    """A decorator indicating abstract properties.

    Requires that the metaclass is ABCMeta or derived from it.  A
    class that has a metaclass derived from ABCMeta cannot be
    instantiated unless all of its abstract properties are overridden.
    The abstract properties can be called using any of the normal
    'super' call mechanisms.

    Usage:

        class C:
            __metaclass__ = ABCMeta
            @abstractproperty
            def my_abstract_property(self):
                ...

    This defines a read-only property; you can also define a read-write
    abstract property using the 'long' form of property declaration:

        class C:
            __metaclass__ = ABCMeta
            def getx(self): ...
            def setx(self, value): ...
            x = abstractproperty(getx, setx)
    """
    __isabstractmethod__ = True


class ABCMeta(type):

    """Metaclass for defining Abstract Base Classes (ABCs).

    Use this metaclass to create an ABC.  An ABC can be subclassed
    directly, and then acts as a mix-in class.  You can also register
    unrelated concrete classes (even built-in classes) and unrelated
    ABCs as 'virtual subclasses' -- these and their descendants will
    be considered subclasses of the registering ABC by the built-in
    issubclass() function, but the registering ABC won't show up in
    their MRO (Method Resolution Order) nor will method
    implementations defined by the registering ABC be callable (not
    even via super()).

    """

    # A global counter that is incremented each time a class is
    # registered as a virtual subclass of anything.  It forces the
    # negative cache to be cleared before its next use.
    _abc_invalidation_counter = 0

    def __new__(mcls, name, bases, namespace):
        cls = super(ABCMeta, mcls).__new__(mcls, name, bases, namespace)
        # Compute set of abstract method names
        abstracts = set(name
                     for name, value in list(namespace.items())
                     if getattr(value, "__isabstractmethod__", False))
        for base in bases:
            for name in getattr(base, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    abstracts.add(name)
        cls.__abstractmethods__ = frozenset(abstracts)
        # Set up inheritance registry
        cls._abc_registry = set()
        cls._abc_cache = set()
        cls._abc_negative_cache = set()
        cls._abc_negative_cache_version = ABCMeta._abc_invalidation_counter
        return cls

    def register(cls, subclass):
        """Register a virtual subclass of an ABC."""
        if not isinstance(subclass, type):
            raise TypeError("Can only register classes")
        if issubclass(subclass, cls):
            return  # Already a subclass
        # Subtle: test for cycles *after* testing for "already a subclass";
        # this means we allow X.register(X) and interpret it as a no-op.
        if issubclass(cls, subclass):
            # This would create a cycle, which is bad for the algorithm below
            raise RuntimeError("Refusing to create an inheritance cycle")
        cls._abc_registry.add(subclass)
        ABCMeta._abc_invalidation_counter += 1  # Invalidate negative cache

    def _dump_registry(cls, file=None):
        """Debug helper to print the ABC registry."""
        print("Class: %s.%s" % (cls.__module__, cls.__name__), file=file)
        print("Inv.counter: %s" % ABCMeta._abc_invalidation_counter, file=file)
        for name in sorted(cls.__dict__.keys()):
            if name.startswith("_abc_"):
                value = getattr(cls, name)
                print("%s: %r" % (name, value), file=file)

    def __instancecheck__(cls, instance):
        """Override for isinstance(instance, cls)."""
        # Inline the cache checking when it's simple.
        subclass = getattr(instance, '__class__', None)
        if subclass in cls._abc_cache:
            return True
        subtype = type(instance)
        # Old-style instances
        if subtype is _InstanceType:
            subtype = subclass
        if subtype is subclass or subclass is None:
            if (cls._abc_negative_cache_version ==
                ABCMeta._abc_invalidation_counter and
                subtype in cls._abc_negative_cache):
                return False
            # Fall back to the subclass check.
            return cls.__subclasscheck__(subtype)
        return (cls.__subclasscheck__(subclass) or
                cls.__subclasscheck__(subtype))

    def __subclasscheck__(cls, subclass):
        """Override for issubclass(subclass, cls)."""
        # Check cache
        if subclass in cls._abc_cache:
            return True
        # Check negative cache; may have to invalidate
        if cls._abc_negative_cache_version < ABCMeta._abc_invalidation_counter:
            # Invalidate the negative cache
            cls._abc_negative_cache = set()
            cls._abc_negative_cache_version = ABCMeta._abc_invalidation_counter
        elif subclass in cls._abc_negative_cache:
            return False
        # Check the subclass hook
        ok = cls.__subclasshook__(subclass)
        if ok is not NotImplemented:
            assert isinstance(ok, bool)
            if ok:
                cls._abc_cache.add(subclass)
            else:
                cls._abc_negative_cache.add(subclass)
            return ok
        # Check if it's a direct subclass
        if cls in getattr(subclass, '__mro__', ()):
            cls._abc_cache.add(subclass)
            return True
        # Check if it's a subclass of a registered class (recursive)
        for rcls in cls._abc_registry:
            if issubclass(subclass, rcls):
                cls._abc_cache.add(subclass)
                return True
        # Check if it's a subclass of a subclass (recursive)
        for scls in cls.__subclasses__():
            if issubclass(subclass, scls):
                cls._abc_cache.add(subclass)
                return True
        # No dice; update negative cache
        cls._abc_negative_cache.add(subclass)
        return False

def _hasattr(C, attr):
    try:
        return any(attr in B.__dict__ for B in C.__mro__)
    except AttributeError:
        # Old-style class
        return hasattr(C, attr)

class Iterable(metaclass=ABCMeta):
    @abstractmethod
    def __iter__(self):
        while False:
            yield None

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Iterable:
            if _hasattr(C, "__iter__"):
                return True
        return NotImplemented

Iterable.register(str)

class Sized(metaclass=ABCMeta):
    @abstractmethod
    def __len__(self):
        return 0

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Sized:
            if _hasattr(C, "__len__"):
                return True
        return NotImplemented


class Container(metaclass=ABCMeta):
    @abstractmethod
    def __contains__(self, x):
        return False

    @classmethod
    def __subclasshook__(cls, C):
        if cls is Container:
            if _hasattr(C, "__contains__"):
                return True
        return NotImplemented

### MAPPINGS ###


class Mapping(Sized, Iterable, Container):

    @abstractmethod
    def __getitem__(self, key):
        raise KeyError

    def get(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            return default

    def __contains__(self, key):
        try:
            self[key]
        except KeyError:
            return False
        else:
            return True

    def iterkeys(self):
        return iter(self)

    def itervalues(self):
        for key in self:
            yield self[key]

    def iteritems(self):
        for key in self:
            yield (key, self[key])

    def keys(self):
        return list(self)

    def items(self):
        return [(key, self[key]) for key in self]

    def values(self):
        return [self[key] for key in self]

    # Mappings are not hashable by default, but subclasses can change this
    __hash__ = None

    def __eq__(self, other):
        if not isinstance(other, Mapping):
            return NotImplemented
        return dict(list(self.items())) == dict(list(other.items()))

    def __ne__(self, other):
        return not (self == other)

class MutableMapping(Mapping):

    @abstractmethod
    def __setitem__(self, key, value):
        raise KeyError

    @abstractmethod
    def __delitem__(self, key):
        raise KeyError

    __marker = object()

    def pop(self, key, default=__marker):
        try:
            value = self[key]
        except KeyError:
            if default is self.__marker:
                raise
            return default
        else:
            del self[key]
            return value

    def popitem(self):
        try:
            key = next(iter(self))
        except StopIteration:
            raise KeyError
        value = self[key]
        del self[key]
        return key, value

    def clear(self):
        try:
            while True:
                self.popitem()
        except KeyError:
            pass

    def update(self, other=(), **kwds):
        if isinstance(other, Mapping):
            for key in other:
                self[key] = other[key]
        elif hasattr(other, "keys"):
            for key in list(other.keys()):
                self[key] = other[key]
        else:
            for key, value in other:
                self[key] = value
        for key, value in list(kwds.items()):
            self[key] = value

    def setdefault(self, key, default=None):
        try:
            return self[key]
        except KeyError:
            self[key] = default
        return default

MutableMapping.register(dict)

################################################################################
### OrderedDict
################################################################################

class OrderedDict(dict, MutableMapping):
    'Dictionary that remembers insertion order'
    # An inherited dict maps keys to values.
    # The inherited dict provides __getitem__, __len__, __contains__, and get.
    # The remaining methods are order-aware.
    # Big-O running times for all methods are the same as for regular dictionaries.

    # The internal self.__map dictionary maps keys to links in a doubly linked list.
    # The circular doubly linked list starts and ends with a sentinel element.
    # The sentinel element never gets deleted (this simplifies the algorithm).
    # Each link is stored as a list of length three:  [PREV, NEXT, KEY].

    def __init__(self, *args, **kwds):
        '''Initialize an ordered dictionary.  Signature is the same as for
        regular dictionaries, but keyword arguments are not recommended
        because their insertion order is arbitrary.

        '''
        if len(args) > 1:
            raise TypeError('expected at most 1 arguments, got %d' % len(args))
        try:
            self.__root
        except AttributeError:
            self.__root = root = [None, None, None]     # sentinel node
            PREV = 0
            NEXT = 1
            root[PREV] = root[NEXT] = root
            self.__map = {}
        self.update(*args, **kwds)

    def __setitem__(self, key, value, PREV=0, NEXT=1, dict_setitem=dict.__setitem__):
        'od.__setitem__(i, y) <==> od[i]=y'
        # Setting a new item creates a new link which goes at the end of the linked
        # list, and the inherited dictionary is updated with the new key/value pair.
        if key not in self:
            root = self.__root
            last = root[PREV]
            last[NEXT] = root[PREV] = self.__map[key] = [last, root, key]
        dict_setitem(self, key, value)

    def __delitem__(self, key, PREV=0, NEXT=1, dict_delitem=dict.__delitem__):
        'od.__delitem__(y) <==> del od[y]'
        # Deleting an existing item uses self.__map to find the link which is
        # then removed by updating the links in the predecessor and successor nodes.
        dict_delitem(self, key)
        link = self.__map.pop(key)
        link_prev = link[PREV]
        link_next = link[NEXT]
        link_prev[NEXT] = link_next
        link_next[PREV] = link_prev

    def __iter__(self, NEXT=1, KEY=2):
        'od.__iter__() <==> iter(od)'
        # Traverse the linked list in order.
        root = self.__root
        curr = root[NEXT]
        while curr is not root:
            yield curr[KEY]
            curr = curr[NEXT]

    def __reversed__(self, PREV=0, KEY=2):
        'od.__reversed__() <==> reversed(od)'
        # Traverse the linked list in reverse order.
        root = self.__root
        curr = root[PREV]
        while curr is not root:
            yield curr[KEY]
            curr = curr[PREV]

    def __reduce__(self):
        'Return state information for pickling'
        items = [[k, self[k]] for k in self]
        tmp = self.__map, self.__root
        del self.__map, self.__root
        inst_dict = vars(self).copy()
        self.__map, self.__root = tmp
        if inst_dict:
            return (self.__class__, (items,), inst_dict)
        return self.__class__, (items,)

    def clear(self):
        'od.clear() -> None.  Remove all items from od.'
        try:
            for node in self.__map.values():
                del node[:]
            self.__root[:] = [self.__root, self.__root, None]
            self.__map.clear()
        except AttributeError:
            pass
        dict.clear(self)

    setdefault = MutableMapping.setdefault
    update = MutableMapping.update
    pop = MutableMapping.pop
    keys = MutableMapping.keys
    values = MutableMapping.values
    items = MutableMapping.items
    iterkeys = MutableMapping.iterkeys
    itervalues = MutableMapping.itervalues
    iteritems = MutableMapping.iteritems
    __ne__ = MutableMapping.__ne__

    def popitem(self, last=True):
        '''od.popitem() -> (k, v), return and remove a (key, value) pair.
        Pairs are returned in LIFO order if last is true or FIFO order if false.

        '''
        if not self:
            raise KeyError('dictionary is empty')
        key = next(reversed(self) if last else iter(self))
        value = self.pop(key)
        return key, value

    def __repr__(self):
        'od.__repr__() <==> repr(od)'
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self.items()))

    def copy(self):
        'od.copy() -> a shallow copy of od'
        return self.__class__(self)

    @classmethod
    def fromkeys(cls, iterable, value=None):
        '''OD.fromkeys(S[, v]) -> New ordered dictionary with keys from S
        and values equal to v (which defaults to None).

        '''
        d = cls()
        for key in iterable:
            d[key] = value
        return d

    def __eq__(self, other):
        '''od.__eq__(y) <==> od==y.  Comparison to another OD is order-sensitive
        while comparison to a regular mapping is order-insensitive.

        '''
        if isinstance(other, OrderedDict):
            return len(self)==len(other) and \
                   all(_imap(_eq, iter(self.items()), iter(other.items())))
        return dict.__eq__(self, other)

    def __del__(self):
        self.clear()                # eliminate cyclical references


################################################################################
### namedtuple
################################################################################

def namedtuple(typename, field_names, verbose=False, rename=False):
    """Returns a new subclass of tuple with named fields.

    >>> Point = namedtuple('Point', 'x y')
    >>> Point.__doc__                   # docstring for the new class
    'Point(x, y)'
    >>> p = Point(11, y=22)             # instantiate with positional args or keywords
    >>> p[0] + p[1]                     # indexable like a plain tuple
    33
    >>> x, y = p                        # unpack like a regular tuple
    >>> x, y
    (11, 22)
    >>> p.x + p.y                       # fields also accessable by name
    33
    >>> d = p._asdict()                 # convert to a dictionary
    >>> d['x']
    11
    >>> Point(**d)                      # convert from a dictionary
    Point(x=11, y=22)
    >>> p._replace(x=100)               # _replace() is like str.replace() but targets named fields
    Point(x=100, y=22)

    """

    # Parse and validate the field names.  Validation serves two purposes,
    # generating informative error messages and preventing template injection attacks.
    if isinstance(field_names, str):
        field_names = field_names.replace(',', ' ').split() # names separated by whitespace and/or commas
    field_names = tuple(map(str, field_names))
    if rename:
        names = list(field_names)
        seen = set()
        for i, name in enumerate(names):
            if (not all(c.isalnum() or c=='_' for c in name) or _iskeyword(name)
                or not name or name[0].isdigit() or name.startswith('_')
                or name in seen):
                names[i] = '_%d' % i
            seen.add(name)
        field_names = tuple(names)
    for name in (typename,) + field_names:
        if not all(c.isalnum() or c=='_' for c in name):
            raise ValueError('Type names and field names can only contain alphanumeric characters and underscores: %r' % name)
        if _iskeyword(name):
            raise ValueError('Type names and field names cannot be a keyword: %r' % name)
        if name[0].isdigit():
            raise ValueError('Type names and field names cannot start with a number: %r' % name)
    seen_names = set()
    for name in field_names:
        if name.startswith('_') and not rename:
            raise ValueError('Field names cannot start with an underscore: %r' % name)
        if name in seen_names:
            raise ValueError('Encountered duplicate field name: %r' % name)
        seen_names.add(name)

    # Create and fill-in the class template
    numfields = len(field_names)
    argtxt = repr(field_names).replace("'", "")[1:-1]   # tuple repr without parens or quotes
    reprtxt = ', '.join('%s=%%r' % name for name in field_names)
    template = '''class %(typename)s(tuple):
        '%(typename)s(%(argtxt)s)' \n
        __slots__ = () \n
        _fields = %(field_names)r \n
        def __new__(_cls, %(argtxt)s):
            return _tuple.__new__(_cls, (%(argtxt)s)) \n
        @classmethod
        def _make(cls, iterable, new=tuple.__new__, len=len):
            'Make a new %(typename)s object from a sequence or iterable'
            result = new(cls, iterable)
            if len(result) != %(numfields)d:
                raise TypeError('Expected %(numfields)d arguments, got %%d' %% len(result))
            return result \n
        def __repr__(self):
            return '%(typename)s(%(reprtxt)s)' %% self \n
        def _asdict(self):
            'Return a new OrderedDict which maps field names to their values'
            return OrderedDict(zip(self._fields, self)) \n
        def _replace(_self, **kwds):
            'Return a new %(typename)s object replacing specified fields with new values'
            result = _self._make(map(kwds.pop, %(field_names)r, _self))
            if kwds:
                raise ValueError('Got unexpected field names: %%r' %% kwds.keys())
            return result \n
        def __getnewargs__(self):
            return tuple(self) \n\n''' % locals()
    for i, name in enumerate(field_names):
        template += '        %s = _property(_itemgetter(%d))\n' % (name, i)
    if verbose:
        print(template)

    # Execute the template string in a temporary namespace and
    # support tracing utilities by setting a value for frame.f_globals['__name__']
    namespace = dict(_itemgetter=_itemgetter, __name__='namedtuple_%s' % typename,
                     OrderedDict=OrderedDict, _property=property, _tuple=tuple)
    try:
        exec(template, namespace)
    except SyntaxError as e:
        raise SyntaxError(e.message + ':\n' + template)
    result = namespace[typename]

    # For pickling to work, the __module__ variable needs to be set to the frame
    # where the named tuple is created.  Bypass this step in enviroments where
    # sys._getframe is not defined (Jython for example) or sys._getframe is not
    # defined for arguments greater than 0 (IronPython).
    try:
        result.__module__ = _sys._getframe(1).f_globals.get('__name__', '__main__')
    except (AttributeError, ValueError):
        pass

    return result



# Command Messaging.
# --
# Copyright 2021 runphase.com .  All rights reserved
#
from .base import CommandMessage, RegisterCommandClass, AsCommandClass # , isTrustedCommand
from . import resource
from . import deployment

from . import subdaemon
from . import shell
from . import python

def ExecuteTrustedCommandMessage(url, payload):
    message = email.message_from_string(payload)
    package = buildCommandFromMessage(message)

    if package is not None:
        if isTrustedCommand(url, package):
            ExecuteCommandMessage(package)

@runtime.api('System::CommandMessaging::API')
class CommandMessagingAPI:
    RegisterCommandClass = staticmethod(RegisterCommandClass)
    ExecuteTrustedCommandMessage = staticmethod(ExecuteTrustedCommandMessage)

# todo: deployment messaging apparatus
# todo: girl script message with player auth



# Miscellaneous MUD Tools.
import types
import linecache
import re

try: import array # cx Freeze
except ModuleNotFoundError: pass

try: import dbm.ndbm, dbm.gnu
except ImportError as e:
    print(f'[dbm import error] {e}')


# Hmm, dependency.  But I don't want stuphlib to depend on this package.
# note: the dependent package is moved into framework/game/format
try: from stuphlib.directory import getUnderAssumedRoot, joinUnderAssumedRoot
except ImportError: pass


def atSetAttr(**kwd):
    def setAttrs(function):
        for (name, value) in kwd.items():
            setattr(function, name, value)

        return function
    return setAttrs


def columnize(items, nr_cols, fmt_width, max_width = None):
    from io import StringIO as new_buffer
    buf = new_buffer()

    if max_width:
        fmt = '%%-%d.%ds' % (fmt_width, max_width)
    else:
        fmt = '%%-%ds' % fmt_width

    col = 1
    nr_cols = int(nr_cols)

    for i in items:
        buf.write(fmt % i)
        if not (col % nr_cols):
            buf.write('\n')

        col += 1

    # XXX no newline added when uneven
    if col % nr_cols:
        buf.write('\n')

    return buf.getvalue()

FORBLANK_PATTERN = re.compile(r'^(\s*)')
def countIndent(line):
    # Match count of whitespace at the beginning of string.
    # Pretty simple, because tab-nontab indentation is incompatible.
    return len(FORBLANK_PATTERN.match(line).group(1))

    ##    i = 0
    ##    for c in line:
    ##        if c in (' ', '\t'):
    ##            i += 1
    ##        else:
    ##            break
    ##
    ##    return i

BLANK_PATTERN = re.compile(r'^(?:\s*)$')
def AllBlank(s):
    return BLANK_PATTERN.match(s) is not None

def ReadFileLines(filename):
    def getline(line):
        if line[-2:] in ['\r\n', '\n\r']:
            return line[:-2]
        if line[-1:] in '\r\n':
            return line[:-1]

        return line

    with open(filename) as fl:
        return list(map(getline, fl))

def functionLines(func, max = 0, trim_decorators = True, width = 60):
    if type(func) not in (types.FunctionType, types.MethodType): #, types.UnboundMethodType):
        raise TypeError(type(func))

    co = func.__code__
    return functionLines0(co.co_filename, co.co_firstlineno - 1,
                          max = max, trim_decorators = trim_decorators,
                          width = width)

def functionLines0(filename, startline, max, trim_decorators, width):
    # The basic source-parsing algorithm detached from function type.
    format = (lambda s, f = '%%.%ds' % width: f % s) if width else (lambda s:s)
    lines = ReadFileLines(filename)

    ln = startline
    end = min(startline + max, len(lines)) if max else len(lines) - 1

    line = lines[ln]
    n = countIndent(line)

    blanks = []
    headerState = 0

    yield format(line)

    while True:
        if ln >= end:
            break

        ln += 1
        line = lines[ln]

        if AllBlank(line):
            blanks.append(line)
            continue

        # The following algorithm only works with proper syntax:
        #   (but doesn't accept one-statement defs on same line) XXX
        i = countIndent(line)
        if i <= n:
            if headerState > 1:
                # Start of next definition.
                break

            if line.lstrip().startswith('@'):
                if trim_decorators:
                    continue

            else:
                # Met beginning of declaration and body.
                headerState = 1

        else:
            # In definition.
            headerState = 2

        if blanks:
            # Flush the blanks cache, but only after we know
            # that we're still within this definition.
            for b in blanks:
                yield format(b)

            blanks = []

        yield format(line)

def functionName(func):
    if type(func) is types.MethodType:
        return '%s.%s.%s' % (func.__module__, func.__self__.__class__.__name__, func.__name__)

    return '%s.%s' % (func.__module__, func.__name__)

def listSortInsert(listObj, item, key = None):
    # Insert an item into a list, with an order in mind.
    # (Maintain priority lists)
    if not listObj:
        # Because the algorithm below can't deal with empty list.
        listObj.append(item)
        return

    if callable(key):
        k = key(item)
        for i in range(len(listObj)):
            if key(listObj[i]) > k:
                break
        else:
            i += 1
    else:
        for i in range(len(listObj)):
            if 0 > cmp(item, listObj[i]):
                break
        else:
            i += 1

    listObj.insert(i, item)

from . import getLineFromCache, clearLineCache, checkLineCache

class lineCache:
    get = staticmethod(getLineFromCache)
    clear = staticmethod(clearLineCache)
    check = staticmethod(checkLineCache)

lineCache = lineCache()

def setupSubmodule(ns, name, base, attrs = ()):
    fqName = name
    if name[0] == '.':
        fqName = ns['__name__'] + name
        name = name[1:]

    import sys
    submod = types.ModuleType(fqName)
    ns[name] = submod
    sys.modules[fqName] = submod

    if isinstance(base, str):
        base = __import__(base, fromlist = [''])

    try: attrs = base.__all__
    except AttributeError: pass

    for n in attrs:
        if isinstance(n, (list, tuple)):
            (n, a) = n
        else:
            a = n

        try: setattr(submod, a, getattr(base, n))
        except AttributeError: pass

    return submod


def xorString(s, k):
    if not k:
        return s

    def forever():
        o = list(map(ord, k))
        while True:
            for c in o:
                yield c

    i = forever().__next__
    return ''.join(chr(ord(c) ^ i()) for c in s)

xor = xorString

def capitalize(string):
    return string[0].upper() + string[1:]

class Attributes(object):
    ##    class Attributes(Object):
    ##        Meta = Object.Meta('function', 'object')

    @classmethod
    def Dynamic(self, function):
        # Decorator.
        def buildGetter(object, *args, **kwd):
            return self(function, object, *args, **kwd)

        return buildGetter

    def __init__(self, function, object, *args, **kwd):
        self.function = function
        self.object = object
        self.args = args
        self.kwd = kwd

    def __getattr__(self, name):
        # Todo: define this policy a bit better.  Also, when to raise AttributeError?
        if name.startswith('__'):
            return object.__getattribute__(self, name)

        return self.function(self.object, name, *self.args, **self.kwd)

def listing(function):
    def listWrapper(*args, **kwd):
        return list(function(*args, **kwd))

    from functools import update_wrapper
    return update_wrapper(listWrapper, function)

def stylizedHeader(caption, nl = '\r\n', blank = True):
    # Display a stuph-styled caption.
    def h():
        yield '&y%s&N' % caption
        yield '&r%s&N' % ('=' * len(caption))

        if blank:
            yield ''

    return nl.join(h())

def expend(sequence):
    # Use up a generator (return nothing)
    for x in sequence:
        pass


class FileNotFoundException(IOError):
    pass

def OpenFile(filename, *args, **kwd):
    # A subroutine for opening a file and raising a separate exception if it doesn't exist.
    # This saves extra code for introspecting the error.
    try: return open(filename, *args, **kwd)
    except IOError:
        from sys import exc_info
        from errno import ENOENT

        (etype, value, tb) = exc_info()

        if value.errno == ENOENT:
            raise FileNotFoundException

        raise etype(value).with_traceback(tb)


# Used..?
def Code(source, name, filename):
    # Compile code.
    tab = '\n' + ' ' * 4

    argspec = '*args, **kwd'
    code = 'def %s(%s):%s%s' % (name, argspec, tab, tab.join(source.split('\n')))

    return compile(code, filename or '<Function Code>', 'exec')

def Function(code, name = None, filename = None, globals = None):
    name = name or 'function'
    if type(code) is str:
        code = Code(code, name, filename)

        # Extract code from function (pre)definition.
        ns = dict()
        exec(code, ns)
        code = ns[name].__code__

    from dis import disassemble as dis
    dis(code)

    name = name or code.co_name

    def mainGlobals():
        import __main__ as main
        return main.__dict__

    import new
    return new.function(code, globals or mainGlobals(), name)

def accessItems(object, names):
    for n in names:
        object = object[n]

    return object


from pickle import dumps
def pickleSafe(object):
    try: return dumps(object)
    except Exception as e:
        return e

# def pickleSafe(object):
#     return object


# todo: put into profiling gc:
##    from cProfile import Profile
##    from mud.runtime.facilities import Facility
##
##    class ProfiledHeartbeatManager(Facility):
##        NAME = 'Heartbeat::Profiler'
##
##        class ProfiledCallable(object):
##            def __init__(self, callable):
##                self.profiler = Profile()
##                self.callable = callable
##            def __call__(self, *args, **kwd):
##                return self.profiler.runcall(self.callable, *args, **kwd)
##            def dump_stats(self, filename):
##                self.profiler.dump_stats(filename)
##
##            def __getattr__(self, name):
##                return object.__getattribute__(self.callable, name)
##
##            @classmethod
##            def isInstance(self, this):
##                return issubclass(this.__class__, self)
##
##        from mud import getBridgeModule
##        getBridgeModule = staticmethod(getBridgeModule)
##
##        @classmethod
##        def create(self):
##            bridgeModule = self.getBridgeModule()
##            heartbeat = getattr(bridgeModule, 'heartbeat', None)
##
##            if not heartbeat:
##                raise Exception('No heartbeat installed!')
##            if self.ProfiledCallable.isInstance(heartbeat):
##                raise Exception('Already installed!')
##
##            bridgeModule.heartbeat = self.ProfiledCallable(heartbeat)
##            return self()
##
##        def __init__(self):
##            self.filename = None
##
##        def __registry_delete__(self):
##            bridgeModule = self.getBridgeModule()
##            heartbeat = getattr(bridgeModule, 'heartbeat', None)
##
##            if not heartbeat:
##                raise SystemError('Heartbeat was uninstalled!')
##            if not self.ProfiledCallable.isInstance(heartbeat):
##                raise SystemError('Expected a profiler installation.')
##
##            bridgeModule.heartbeat = heartbeat.callable
##
##        class Manager(Facility.Manager):
##            VERB_NAME = 'heartbeat-*profiler'
##            MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR
##
##            def do_filename(self, peer, cmd, args):
##                facility = self.facility.get()
##                if len(args) == 0:
##                    print >> peer, facility.filename
##                else:
##                    facility.filename = ' '.join(args)
##
##            def do_dump(self, peer, cmd, args):
##                facility = self.facility.get()
##                if not facility.filename:
##                    print >> peer, 'Filename not specified.'
##
##                heartbeat = getattr(facility.getBridgeModule(), 'heartbeat', None)
##                if not heartbeat:
##                    raise SystemError('Heartbeat was uninstalled!')
##                if not self.facility.ProfiledCallable.isInstance(heartbeat):
##                    raise SystemError('Expected a profiler installation.')
##
##                heartbeat.profiler.dump_stats(facility.filename)
##
##            def do_list(self, peer, cmd, args):
##                facility = self.facility.get()
##                if not facility.filename:
##                    print >> peer, 'Filename not specified.'
##
##                heartbeat = getattr(facility.getBridgeModule(), 'heartbeat', None)
##                if not heartbeat:
##                    raise SystemError('Heartbeat was uninstalled!')
##                if not self.facility.ProfiledCallable.isInstance(heartbeat):
##                    raise SystemError('Expected a profiler installation.')
##
##                from cStringIO import StringIO
##                from pstats import Stats
##
##                args = list(args)
##                args.insert(0, 'list')
##                (options, args) = self.parse_list_args(args)
##
##                if options.output_file:
##                    buf = open(options.output_file, 'w')
##                else:
##                    buf = StringIO()
##
##                stat = Stats(facility.filename, stream = buf)
##
##                stat.strip_dirs()
##                # stat.sort_stats(...)
##
##                for arg in args:
##                    if arg == 'callers':
##                        stat.print_callers()
##                    elif arg == 'callees':
##                        stat.print_callees()
##                    elif arg == 'stats':
##                        stat.print_stats()
##
##                if options.output_file:
##                    buf.flush()
##                    buf.close()
##                else:
##                    peer.page_string(buf.getvalue())
##
##            def parse_list_args(self, args):
##                from optparse import OptionParser
##                parser = OptionParser()
##                parser.add_option('--output-file')
##                return parser.parse_args(args)

# ProfiledHeartbeatManager.manage()



# Player Interpreter -- Code Evaluators Stack.
# Copyright 2011 Clint Banis.  All rights reserved.
#
from stuphos.etc.tools import columnize, checkLineCache, parseOptionsOverSystem
from stuphos.runtime import Object, LookupObject as LookupFactoryClass
from stuphos.runtime import LookupObject as LookupProgrammeManagerObject
from stuphmud.server.player import isFromSecureHost

# Record compiled statements to file so that tracebacks can look up the source code.
USE_RECORD_COMPILER = True
from os import unlink as delete_file
from optparse import OptionParser

class CodeManager:
    # This kind of functionality could every well just go into CommandManager
    def __init__(self, *evaluators):
        self.evaluators = list(evaluators)

    def pushEvaluator(self, e):
        self.evaluators.append(e)
    def popEvaluator(self):
        return self.evaluators.pop()

    def getCurrentEvaluator(self):
        try: return self.evaluators[-1]
        except IndexError: pass

    def evaluateStatement(self, peer, argstr):
        evaluate = self.getCurrentEvaluator()
        if evaluate is not None:
            return evaluate(self, peer, argstr)

    def getEvaluatorByManagerName(self, name):
        # First in reverse.
        v = self.evaluators
        i = len(v) - 1
        while i >= 0:
            e = v[i]
            if e.manager.getManagerName() == name:
                return e

            i -= 1

class Programme(Object):
    class _Meta(Object._Meta):
        Attributes = Object._Meta.Attributes + ['manager', 'name', 'codeSource', 'owner']

    # Shelvable.
    def __init__(self, manager, name, sourceCode,
                 owner = None, codeSource = None,
                 documentation = ''):

        self.manager = manager
        self.name = name
        self.sourceCode = sourceCode
        self.owner = owner
        self.codeSource = codeSource
        self.documentation = documentation

    def __getstate__(self):
        # Basically, the manager instance 'goes away' when shelved.
        return dict(name = self.name, owner = self.owner,
                    codeSource = self.codeSource,
                    sourceCode = self.sourceCode,
                    documentation = self.documentation,
                    managerId = self.getManagerId())

    def getManagerId(self):
        # Called during serialization.
        try: return self.managerId
        except AttributeError:
            managerId = self.managerId = self.getManager().getManagerId()
            return managerId

    def getManager(self):
        try: return self.manager
        except AttributeError:
            # Called during invocation/deserialization.
            manager = self.manager = LookupProgrammeManagerObject(self.managerId)
            return manager

    def getShortName(self):
        return self.name
    def getDisplayForm(self):
        manager = self.getManager().getManagerName()

        doc = self.documentation
        if doc:
            doc = '\n  '.join(doc.split('\n'))
            return '[ %s: %s ]\n  %s\n\n%s' % (manager, self.name, doc, self.sourceCode)

        return '[ %s: %s ]\n\n%s' % (manager, self.name, self.sourceCode)

    def __str__(self):
        return self.sourceCode

    # ProgrammeManager Implementation.
    def invokeProgramme(self, shell, peer, **environ):
        # By default, ProgrammeManagers need to define compileSourceCodeBlock and executeCode.
        manager = self.getManager()
        code = manager.compileSourceCodeBlock(peer, self.sourceCode)
        manager.executeCode(shell, peer, code, **environ)

# How to make programs notebook-based?  Well, it's keyed on avatar.programs object (which should
# look like a shelf, so technically compatability already exists.  Of course, most other notebooks
# are only open long enough to do one operation.)
class ProgrammeManager: # (mud.player.notebook.Notebook.Model)
    # Shelving.
    def getPlayerProgrammes(self, peer):
        try: return peer.avatar.programs
        except AttributeError:
            pass

    def getManagerName(self):
        return '--'
    def getManagerId(self):
        raise NotImplementedError

    def formatSourceCode(self, source):
        return source.replace('\r', '')

    ProgrammeClass = Programme

    def buildProgramme(self, peer, name, source):
        # Override this.
        return self.ProgrammeClass(self, name, self.formatSourceCode(source))
                                   # owner = peer

    def loadProgramme(self, peer, name):
        'Load a program off avatar using dictionary mapping.'

        try: return self.getPlayerProgrammes(peer)[name]
        except (TypeError, KeyError):
            pass

        return ''

    def saveProgramme(self, peer, name, source):
        'Save a program onto avatar using dictionary mapping.'

        programs = self.getPlayerProgrammes(peer)
        if programs is not None:
            try:
                programs[name] = source
                programs.sync()

            except (TypeError, AttributeError):
                pass

    def showProgrammes(self, peer, args):
        'Show programmes available through avatar.'

        programs = self.getPlayerProgrammes(peer)
        if programs is not None:
            if args:
                name = ' '.join(args)

                try: p = programs[name]
                except (TypeError, KeyError): pass
                else:
                    if isinstance(p, Programme):
                        p = p.getDisplayForm()
                    else:
                        p = str(p)

                    peer.page_string(p)
            else:
                programs = list(programs.items())
                if programs:
                    def getProgramDescriptions():
                        for (name, p) in programs:
                            if isinstance(p, Programme):
                                yield p.getShortName()
                            else:
                                yield str(name)

                    peer.page_string('&yPrograms&N\r\n&r========&N\r\n%s\r\n' % \
                                     columnize(getProgramDescriptions(), 2, 20))
                else:
                    print('You have no programs.', file=peer)

    def deleteProgramme(self, peer, name):
        'Delete programme, as named, available through avatar.'

        if name:
            programs = self.getPlayerProgrammes(peer)
            if programs is not None:
                try:
                    del programs[name]
                    programs.sync()

                except (TypeError, KeyError): pass
                else:
                    return True

class EvaluationShell:
    # Utilities for managing and executing code.
    def __init__(self, manager):
        self.manager = manager

    def executeSourceCode(self, shell, peer, argstr):
        # argstr = delete_doubledollar(argstr)

        if self.isCodeBlock(argstr):
            self.executeCodeBlock(shell, peer, argstr[0], argstr[1:])
        else:
            self.executeSingleStatement(shell, peer, argstr)

        return True

    def isCodeBlock(self, argstr):
        # todo: something cooler like:
        # ;trigger=enter-game
        # ;trigger+=enter-game
        # ;!trigger:enter-game

        return argstr[0] in '=+;!'

    def executeDirective(self, peer, args):
        # Directive mode.
        larg = args[0].lower() if args else ''

        if larg in ('list', 'show'):
            self.manager.showProgrammes(peer, args[1:])
            return True

        elif larg == 'help':
            print(';; - Execute directive:', file=peer)
            print('     list/show help delete', file=peer)
            print('', file=peer)
            print('=<name> - Save code block to programme named', file=peer)
            print('+<name> - Load, edit and save code block as named', file=peer)
            print('!<name> - Executed named programme', file=peer)
            print('', file=peer)
            return True

        elif larg == 'delete':
            name = ' '.join(args[1:])
            if self.manager.deleteProgramme(peer, name):
                print('%r deleted.' % name, file=peer)

            return True

        elif args:
            print('Unknown code block directive: %r' % ' '.join(args), file=peer)
            return True

    PROGRAMMING_HEADER = '[ %s ] &B*&N ---------------               --------------- &B*&N [ Edit ]'

    def startEditingCodeBlock(self, shell, peer, block, name = '', initialSource = ''):
        # Dispatch the programmed code block by deferring evaluation and entering string-edit.
        def dispatchProgramWithPeerHead(peer, source):
            # Cleanup.
            try: del peer.messenger
            except AttributeError:
                pass # peer.messenger = None

            return shell.withPeerHead(peer, block, source)

        # Start editing.
        peer.messenger = dispatchProgramWithPeerHead
        peer.editString(initialSource)

        # Complete the line in case the peer is in compact mode.
        name = (name and name.upper() or 'Program')
        print(self.PROGRAMMING_HEADER % name, file=peer) # ,

    def parseProgramName(self, argstr):
        parts = argstr.split(' ', 1)
        if len(parts) == 2:
            return parts
        elif len(parts) == 1:
            return (parts[0], '')
        return ('', '')

    def executeCodeBlock(self, shell, peer, mode, argstr):
        # Handle code block mode.
        if mode == ';':
            if not self.executeDirective(peer, argstr.split()):
                # Basic batch compile and execute (with existing source).
                def CompileAndExecute(peer, source):
                    # Direct compile and execute.
                    source = self.manager.formatSourceCode(source)
                    self.executeCode(shell, peer, self.compileSourceCodeBlock(peer, source))

                self.startEditingCodeBlock(shell, peer, CompileAndExecute)

        elif mode == '=':
            # Edit and save programme, replacing old.
            (name, rest) = self.parseProgramName(argstr)
            if not name:
                print('A programme name is required.', file=peer)
            else:
                def StoreProgramme(peer, source):
                    # (should be passing this shell)
                    programme = self.manager.buildProgramme(peer, name, source)
                    self.manager.saveProgramme(peer, name, programme)

                self.startEditingCodeBlock(shell, peer, StoreProgramme, name = name)

        elif mode == '+':
            # Load, edit and save existing programme.
            (name, rest) = self.parseProgramName(argstr)
            if not name:
                print('A programme name is required.', file=peer)
            else:
                programme = self.manager.loadProgramme(peer, name)
                if isinstance(programme, Programme):
                    initialSource = programme.sourceCode
                elif isinstance(programme, str):
                    initialSource = programme
                else:
                    initialSource = ''

                def StoreProgramme(peer, source):
                    # (should be passing this shell)
                    programme = self.manager.buildProgramme(peer, name, source)
                    self.manager.saveProgramme(peer, name, programme)

                self.startEditingCodeBlock(shell, peer, StoreProgramme, name = name,
                                           initialSource = initialSource)

        elif mode == '!':
            # Load programme.
            (name, rest) = self.parseProgramName(argstr)
            if not name:
                print('A programme name is required.', file=peer)
            else:
                # Execute right now.
                programme = self.manager.loadProgramme(peer, name)
                if isinstance(programme, Programme):
                    programme.invokeProgramme(shell, peer, argstr = rest)
                elif isinstance(programme, str):
                    code = self.compileSourceCodeBlock(peer, programme)
                    self.executeCode(shell, peer, code, argstr = rest)

    def executeSingleStatement(self, shell, peer, source):
        self.executeCode(shell, peer, self.compileSingleStatement(peer, source))

    # Abstractions.
    ##    def executeCode(self, shell, peer, code): pass
    ##    def executeSingleStatement(self, shell, peer, source): pass
    ##    def compileSourceCodeBlock(self, peer, source): pass
    ##    def compileSingleStatement(self, peer, sourceCode, type = 'single', ext = ''): pass

# Todo: move this kind of thing into mud.tools?
class TemporaryCodeFile:
    'As context -- Compile to temporary file and return the code.  Delete on exit.'

    def __init__(self, peer, sourceCode, ext = '', shellName = '-'):
        self.peer = peer
        self.sourceCode = sourceCode
        self.ext = ext
        self.shellName = shellName

    def shouldUse(self):
        return getattr(self.peer, 'recordCompiler', USE_RECORD_COMPILER)

    def __enter__(self):
        if self.shouldUse():
            try:
                # Build file name -- this should go in plrfiles.
                peer = self.peer
                filename = 'etc/code/%s@%s%s' %      \
                        (peer.avatar.name            \
                           if (peer.avatar and       \
                               peer.avatar.isPlayer) \
                           else '-', self.shellName, self.ext)

                # Do this first.
                checkLineCache(filename)

                # Write to file.
                fl = open(filename, 'wt')
                fl.write(self.sourceCode)
                fl.close()

            except:
                pass
            else:
                self.filename = filename

        return self

    def __exit__(self, etype = None, value = None, tb = None):
        try: delete_file(self.filename)
        except AttributeError:
            pass

# Factory routines.
class ShellFactoryBase(object):
    @classmethod
    def parseArgs(self, args):
        try: parser = self.__parser
        except AttributeError:
            try: options = self.OPTIONS
            except AttributeError: parser = None
            else:
                parser = self.__parser = OptionParser()
                for (opt_args, opt_kwd) in options:
                    parser.add_option(*opt_args, **opt_kwd)

        if parser is not None:
            try: (options, args) = parseOptionsOverSystem(parser, list(args))
            except SystemExit: return
            else: return options.__dict__

        return dict()

from .python import getPythonCodeEvaluator
try: from .girl import getGirlCodeEvaluator
except ImportError as e:
    # from stuphos.system.api import syslog
    # syslog('[player.interfaces.code.girl.evaluator] %s: %s' % (e.__class__.__name__, e))
    pass

def makeCodeEvaluators(*args, **kwd):
    return [getPythonCodeEvaluator(*args, **kwd)]

# Todo: Initialize this from MUDConf.
SHELL_FACTORIES = dict(girl = 'mud.player.interfaces.code.girl.GirlCodeEvaluatorFactory',
                       agent = 'mud.player.interfaces.code.girl.GirlCodeEvaluatorFactory',
                       python = 'mud.player.interfaces.code.python.PythonCodeEvaluatorFactory')

def findShellFactory(name):
    try: cls = SHELL_FACTORIES[name]
    except KeyError: pass
    else: return LookupFactoryClass(cls)

def doManageShellEvaluators(peer, action, name = None, *args):
    if action == 'open':
        factory = findShellFactory(name)
        if factory is None:
            print('Unknown shell: %r' % name, file=peer)
        else:
            options = factory.parseArgs(args)
            if isinstance(options, dict):
                # Open a shell context.
                try: pushEv = peer.interpreter.pushEvaluator
                except AttributeError:
                    print('Your shell interpreter does not support new evaluators.', file=peer)
                else:
                    pushEv(factory(peer, **options))

    elif action == 'close':
        # Close this shell context.
        try: popEv = peer.interpreter.popEvaluator
        except AttributeError:
            print('Your shell interpreter does not support evaluators.', file=peer)
        else:
            manager = peer.interpreter.getCurrentEvaluator().manager
            print('Closing Program: %s.' % manager.getManagerName(), file=peer)
            popEv()

    elif action == 'show':
        try: evaluators = peer.interpreter.evaluators
        except AttributeError:
            print('Your shell interpreter does not support evaluators.', file=peer)
        else:
            peer.page('\n'.join(map(repr, evaluators)) + '\n')



# StuphOS Compartmentalized Django WSGI Development/Production Server
#   for Django-1.2.1

# This contains a django framework and a bit of an application framework
# for using as a component to building applications.  But, the applications
# are specific to embedding contexts (usually), or, just other purposes
# that are fulfilled by the role of this software package.  However, these
# things are more network specific (external interfaces like the web).

# It is necessary to first import this module to setup the django settings
# module before using the server in this package.  Fortunately, this happens
# automatically.


from django.core.servers.basehttp import run, WSGIRequestHandler, WSGIServer, ServerHandler
#from django.core.servers.basehttp import WSGIServerException
#from django.core.servers.basehttp import AdminMediaHandler
#from django.core.management.validation import get_validation_errors
from django.core.handlers.wsgi import WSGIHandler
from django.conf import settings
from django.utils import translation
from django.core.servers.basehttp import WSGIRequestHandler
from django.utils.deprecation import MiddlewareMixin

from stuphos.runtime.facilities import Facility
from stuphos.runtime.registry import getObject
from stuphos.runtime import newComponent
from stuphos.runtime.architecture.api import writeprotected

from stuphos import getConfig
from stuphos.etc import isYesValue

# Done dynamically:
# from stuphos.structure import RequestAdapter

from io import StringIO, BytesIO
from _thread import start_new_thread as nth
from threading import Thread
from select import error as select_error
from errno import EINTR
from contextlib import contextmanager
from socketserver import BaseServer

from http.client import HTTPResponse as baseHTTPResponse

import os
import sys
import imp


DATA_RETENTION_KEY = False # Experimental

# Make Django project and webapps visible.
#

_settings_module = '%s.project.settings' % __name__
if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = _settings_module

def importDjango(*submodules):
    import django
    return django


# Server Model Validation.
class ValidationError(Exception):
    pass

def ValidactivateOnce(app = None):
    if not getattr(sys, '__django_validactivated', False):
        ##    s = StringIO()
        ##    num_errors = get_validation_errors(s, app)
        ##    if num_errors:
        ##        raise ValidationError("One or more models did not validate:\n%s" % s.getvalue())

        #translation.activate(settings.LANGUAGE_CODE)
        sys.__django_validactivated = True

def getWSGIServerExceptionMessage(e):
    # Use helpful error messages instead of ugly tracebacks.
    ERRORS = {
        13: "You don't have permission to access that port.",
        98: "That port is already in use.",
        99: "That IP address can't be assigned-to.",
    }
    try: return ERRORS[e.args[0].args[0]]
    except (AttributeError, KeyError):
        return str(e)

try: from fcntl import fcntl, F_SETFD, F_GETFD, FD_CLOEXEC
except ImportError: CloseOnExec = False
else:
    def CloseOnExec(fd):
        # My Impl:
        ##    'This is broken because ~1 returns -2!  Fortunately, sockets seem to be !FD_CLOEXEC initially.'
        ##    oldbits=fcntl(fd, F_GETFD)
        ##    newbits=oldbits & ~FD_CLOEXEC
        ##
        ##    #print oldbits, '&', '~%d' % FD_CLOEXEC, '(%d)' % ~FD_CLOEXEC, '=', newbits
        ##
        ##    ##fcntl(fd, F_SETFD, newbits)
        ##    return fd

        # Basically, from solution addressing [Bug #1222790]:
        flags = fcntl(fd, F_GETFD)
        flags |= FD_CLOEXEC
        fcntl(fd, F_SETFD, flags)

# Service.
from socketserver import ThreadingMixIn as ServerCompartment
class ServerControl(ServerCompartment):
    def set_running(self, value = True):
        self.__running = value
    def is_running(self):
        try: return (self.__running)
        except AttributeError:
            return False

    ##    def process_request_thread(self, *args, **kwd):
    ##        import pdb; pdb.set_trace()
    ##        return ThreadingMixin.process_request_thread(self, *args, **kwd)

    def server_start(self):
        def serve_loop():
            # import pdb; pdb.set_trace()

            while self.is_running():
                #import pdb; pdb.set_trace()

                # print('handling django request...')
                # debugOn()

                try: self.handle_request()
                except select_error as e:
                    if e.args[0] is EINTR:
                        # A process signal was sent, ignore and continue.
                        continue

                    (etype, value, tb) = sys.exc_info()
                    raise etype(value).with_traceback(tb)
                # I don't know where WSGIServerException is defined in django lib.
                # except WSGIServerException, e:
                #     self.log(getWSGIServerExceptionMessage(e))
                except KeyboardInterrupt:
                    self.server_stop()

        self.set_running(True)
        Thread(target = serve_loop, daemon = True).start()
        # nth(serve_loop, ())

    def server_stop(self):
        self.set_running(False)
    def server_shutdown(self):
        # XXX!
        self.server_stop()
        self.server_close()

    def runningState(self):
        return self.is_running() and 'RUNNING' or 'STOPPED'

    def handle_error(self, request, client_address):
        from stuphos.system.api import syslog
        from stuphos import logException, getConfig

        from sys import exc_info
        from socket import error
        from errno import EPIPE
        (etype, value, tb) = exc_info()

        def log():
            syslog('[HTTP %s] %s: %s [%s]' % ('%s:%s' % client_address, etype.__name__, value, request))

            if isinstance(value, error):
                if value.errno == EPIPE:
                    return

            logException(traceback = True, header = False)

        # method = getConfig('process-error') # XXX :skip: currently disabled because invocation not called right, 'core/system/httpError')
        # if method:
        #     from stuphos.kernel import invokeLibraryMethod as i
        #     task = i(method, etype, value, tb,
        #              request, client_address)

        #     @nth
        #     def asynchronous():
        #         # Why does this need to be off-thread again??
        #         # I ask because handle_error should already be parallel to VM thread.

        #         # XXX :skip: This is also wrong because the task variable is actually bound
        #         #   to a result object from a task that has run.

        #         task = task()

        #         @task.onComplete
        #         def completion(task, exception = None, **kwd):
        #             log()

        # else:
        #     log()


class Instrument:
    # These models are not currently available as named:
    ##    from .tracking import onSendMail
    ##    from .tracking import onMovement
    ##    from .tracking import onDealDamage
    ##    from .tracking import onSlayMobile
    ##    from .tracking import onDeathTrap
    ##    from .tracking import onPasswordChange
    ##    from .tracking import onPlayerSavePoint
    pass

class EmbeddedRequestHandler(WSGIRequestHandler):
    # Provide configurable ServerHandler.server_software
    # (Probably a better way to do this is to set Server header)

    class ServerHandler(ServerHandler):
        def __init__(self, stdin, stdout, stderr, environ, software, **kwd):
                     # multithread = True, multiprocess = False):
            ServerHandler.__init__(self, stdin, stdout, stderr, environ, **kwd)
            self.server_software = software

        def setup_environ(self):
            ServerHandler.setup_environ(self)
            self.environ['stuphos.streaming.handler'] = self

        def close(self):
            if configurationTruth.DjangoService.log_real_ip:
                try: self.request_handler.x_real_ip = self.environ['HTTP_X_REAL_IP']
                except KeyError: pass
                # Todo: get rid of this once we know it works.
                # else: print(f'[x-real-ip] {self.request_handler.x_real_ip} -> {self.request_handler.client_address[0]}')

            return ServerHandler.close(self)


    def handle(self):
        """Copy of WSGIRequestHandler, but with different ServerHandler"""

        # [HTTP 192.168.43.1:52786] ConnectionResetError: [Errno 104] Connection reset by peer [<ssl.SSLSocket fd=15, family=AddressFamily.AF_INET, type=SocketKind.SOCK_STREAM, proto=0, laddr=('192.168.43.192', 2180)>]
        #   [socketserver.py:654] process_request_thread
        #     self.finish_request(request, client_address)
        #   [socketserver.py:364] finish_request
        #     self.RequestHandlerClass(request, client_address, self)
        #   [socketserver.py:724] __init__
        #     self.handle()
        #   [stuphos/webserver/__init__.py:211] handle
        #     self.raw_requestline = self.rfile.readline(65537)
        #   [socket.py:586] readinto
        #     return self._sock.recv_into(b)
        #   [ssl.py:971] recv_into
        #     return self.read(nbytes, buffer)
        #   [ssl.py:833] read
        #     return self._sslobj.read(len, buffer)
        #   [ssl.py:590] read
        #     v = self._sslobj.read(len, buffer)

        self.raw_requestline = self.rfile.readline(65537)
        if len(self.raw_requestline) > 65536:
            self.requestline = ''
            self.request_version = ''
            self.command = ''
            self.send_error(414)
            return

        if not self.parse_request():  # An error code has been sent, just exit
            return

        handler = self.ServerHandler(
            self.rfile, self.wfile, self.get_stderr(), self.get_environ(),
            self.server.software
        )
        handler.request_handler = self      # backpointer for logging
        # self.handler = handler # for streaming
        handler.run(self.server.get_app())

class LoggingWSGIRequestHandler(EmbeddedRequestHandler):
    'Disables logging of WSGI messages (conditionally).'

    def log_message(self, format, *args):
        # import pdb; pdb.set_trace()

        # XXX where is this coming from?
        # if "Request timed out: timeout('The read operation timed out')" in str(format):
        # if "Request timed out" in str(format):
        #     debugOn()

        # debugOn()

        if isYesValue(getConfig('log-request', 'DjangoService')):
            message = '%s - - [%s] %s' % \
                         (getattr(self, 'x_real_ip', self.client_address[0]),
                          self.log_date_time_string(),
                          format)

            EmbeddedRequestHandler.log_message(self, message, *args)

# from django.utils.deprecation import MiddlewareMixin

class Key:
    public = ''
    private = ''

    def __init__(self, request = None):
        if request:
            try:
                # (Never encrypt)
                self.private = request.session.get('retention_token_secret', self.private)
                self.public = request.session.get('retention_token_public', self.public)
 
            except KeyError:
                pass

            else:
                return

        # pair = providerGenerate()
        # self.private = pair.private
        # self.public = pair.public

    def __call__(self, data):
        return data # encrypt(data, self.public)


class LoggingMiddleware(MiddlewareMixin):
    # Log to journal (database)
    @runtime.available(runtime.System.Journal)
    # Note the argument positioning.
    def process_view(log, self, request, callback, callback_args, callback_kwargs):
        from stuphos.etc.tools.timing import date as now
        import json

        # todo: debug this api:
        # @runtime.System.Engine # (accessorOf.contextObject.task)
        # def task(vm):
        #     return vm.contextObject.task

        task = None

        source = task.taskName if task is not None else 'web-request'
        meta = request.META.get

        # print(str(dict(request.META).keys()))

        # session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME)

        content = dict(remoteAddress = meta('REMOTE_ADDR'),
                       realIPAddress = meta('HTTP_X_REAL_IP'),
                       userAgent = meta('HTTP_USER_AGENT'),
                       referer = meta('HTTP_REFERER'),
                       method = meta('REQUEST_METHOD'),
                       path = meta('PATH_INFO'),
                       queryString = meta('QUERY_STRING'))

        # USER INFO: session key is temporary whereas username is permanent.
        # if session_key:
        #     # User info.
        #     content['sessionKey'] = session_key

        content = json.dumps(content, indent = 1)

        if DATA_RETENTION_KEY:
            key = Key() # (request)
            content = json.dumps(dict(content = key(content), user = request.user.id))

            request.session['retention_token_secret'] = key.private
            request.session['retention_token_public'] = key.public

        # debugOn()
        # :system-journal:
        log += dict(source = source, type = 'web-request/json',
                    content = content, timestamp = now())


class MultiTenancy(MiddlewareMixin):
    def process_view(log, self, request, callback, callback_args, callback_kwargs):
        from django.conf import settings

        rewrite = configuration.DjangoService.tenancy_rewrite_format \
                    or 'tenancy/{tenancy}/{path}'

        try: tenancy = settings.MULTITENANCY[request.META['HTTP_HOST']]
        except (AttributeError, KeyError): pass
        else: request.path = rewrite.format(tenancy = tenancy, path = path)


# class CsrfExemptViewMiddleware(MiddlewareMixin):
#     def process_request(self, request):
#         if self.turnOffCsrf(request):
#             request._dont_enforce_csrf_checks = True

#     def turnOffCsrf(self, request):
#         return request.path in getattr(settings, 'CSRF_EXEMPT_PATHS', [])


@runtime.available(runtime.System.Journal)
def retentionView(log, request, key = False, delete = False):
    # io.user.data.jsonOf.write(host['/data/retention'].value)
    # host['/data/retention?delete=true'].value 
    # io.user.secret.keyOf.write(host['/data/retention?key=true'].value)
    # host['/data/retention?key=true&delete=true'].value

    key = request.GET.get('key', key)

    def records(self):
        for e in log.select():
            # todo: configure task context in middleware
            if e['source'] != 'web-request':
                continue

            o = json.loads(e.content)

            if o['user'] == request.user.id:
                yield (e, o['content'])

                if delete:
                    e.delete()

    # todo: assert SSL, overlock
    if key:
        if key not in ['yes', 'true', '1', 'on', 'no', 'false', '0', 'off']:
            # Set key.
            assert not any(records())

            del request.session['retention_token_public']
            request.session['retention_token_secret'] = key

            key = False

        else:
            key = isYesValue(key)

            if key:
                o = request.session['retention_token_secret']
                response = http.DjangoResponse(o, content_type = 'text/plain') # application/octet

    delete = isYesValue(request.GET.get('delete', delete) or 'no')

    if key:
        # Transactionally return key on confirmation of exported records.

        if delete:
            del request.session['retention_token_secret']
            del request.session['retention_token_public']

        return response

    return http.DjangoResponse(json.dumps(list(records()), indent = 1),
                               content_type = 'application/json')


class DjangoService(Facility, ServerControl, WSGIServer, Instrument):
    NAME = 'Django::Service'

    # Configuration.
    HOSTNAME = '0.0.0.0' # 'localhost'
    PORT = 8000
    ADMIN_MEDIA_PATH = ''
    DEBUG = False # Production

    @classmethod
    def GetConfig(self):
        try:
            from stuphos import getConfig
            from stuphos.etc import isYesValue

        except ImportError:
            hostname = self.HOSTNAME
            port = self.PORT
            admin_media_path = self.ADMIN_MEDIA_PATH
            debug = self.DEBUG

            certfile = None
            keyfile = None
            software = ServerHandler.server_software

            urlConf = 'phsite.network.urls' # XXX default this better

        else:
            hostname = getConfig('hostname', 'DjangoService') or self.HOSTNAME
            port = int(getConfig('port', 'DjangoService') or self.PORT)
            admin_media_path = getConfig('admin-media-path', 'DjangoService') or self.ADMIN_MEDIA_PATH
            debug = self.DEBUG # isYesValue(getConfig('show-debug-page', 'DjangoService') or (self.DEBUG and 'yes' or 'no'))

            certfile = getConfig('certificate-path', 'DjangoService')
            keyfile = getConfig('keyfile-path', 'DjangoService')
            software = getConfig('software', 'DjangoService')

            urlConf = getConfig('sitemap', 'DjangoService')

        return (hostname, port, admin_media_path, debug,
                certfile, keyfile, software, urlConf)

    class Manager(Facility.Manager):
        VERB_NAME = 'dj*ango-service'
        MINIMUM_LEVEL = Facility.Manager.IMPLEMENTOR

        def do_reload(self, peer, parentCommand, args):
            with ReloadEmbeddedUrlConf():
                for moduleName in args:
                    try: module = importModule(moduleName)
                    except ImportError as e: print('%s: %s' % (moduleName, str(e)), file=peer)
                    else: print('Reloaded: %s' % getattr(imp.reload(module), '__name__', moduleName), file=peer)

            print('Reloaded Django URLConf.', file=peer)

        def do_toggle_debug(self, peer, command, args):
            from django.conf import settings
            settings.DEBUG = not settings.DEBUG

    # Uncomment this to debug individual incoming requests:
    # finish_request = breakOn(WSGIServer.finish_request)

    def server_bind(self, *args, **kwd):
        # Ideally, this would go on ServerControl, if WSGIServer/BaseHTTPServer actually implemented it.
        result = WSGIServer.server_bind(self, *args, **kwd)
        if CloseOnExec:
            CloseOnExec(self.fileno())

        return result

    @classmethod
    def create(self):
        from phsite.network.embedded import install, uninstall
        install()

        import logging, logging.handlers
        logger = logging.getLogger('django.server')
        logger.addHandler(logging.StreamHandler())
        logger.setLevel(logging.INFO)

        try:
            (hostname, port, admin_media_path, debug,
             certfile, keyfile, software, urlConf) = \
                self.GetConfig()

            djangoConfig()

            #wsgi_handler = AdminMediaHandler(WSGIHandler(), admin_media_path)
            wsgi_handler = WSGIHandler()

            server = newComponent(self) # todo: ditch instrument, use Module component.
            self.__init__(server, (hostname, port), wsgi_handler,
                          urlConf, certfile, keyfile, software)
        except:
            uninstall()
            raise

        return server

    def __init__(self, server_address, wsgi_handler, urlConf, certfile, keyfile, software):
        WSGIServer.__init__(self, server_address, LoggingWSGIRequestHandler)
        self.software = software

        if certfile is not None:
            import ssl

            kwd = dict(certfile = certfile)

            if keyfile is not None:
                kwd['keyfile'] = keyfile

            self.socket = ssl.wrap_socket (self.socket, server_side = True, **kwd)

        self.set_app(wsgi_handler)

        # Start Immediately.
        self.server_start()

        # Force load the sitemap. XXX :skip: Hack
        ##    from django.core.urlresolvers import get_urlconf, get_resolver
        ##    urlconf = get_urlconf()
        ##    resolver = get_resolver(urlconf)
        ##    resolver._populate()


        # Configuration specific for the development shell.
        # Note:
        # django\core\urlresolvers.py(107) #get_resolver = memoize(get_resolver, _resolver_cache, 1)

        ##    DATABASE_PATH = str(io.here.web.olc.dbOf)
        ##    DATABASE = dict(ENGINE = 'django.db.backends.sqlite3',
        ##                    NAME = DATABASE_PATH,
        ##                    USER = '',
        ##                    PASSWORD = '')

        #settings._wrapped = None # This seems to work, although it's hardcore.
        #settings.configure(DATABASES = dict(default = DATABASE))

        # settings.DATABASES = dict(default = DATABASE)

        # Do this.. the problem is that the settings are already all
        # configured way before, and when the database modules are
        # initialized, because django is imported initially by pywin
        # shell.  So... for the purposes of this testing, we'd need
        # to reload alot of the components that require 'connections'
        # OR, find a way to configure the database manually within
        # pywin.
        #import django
        #reload(django.db)

        # OR, figure out a way to host multiple databases/settings.

        # This was moved to before server/handler construction (create()).
        ##    from django.apps import apps
        ##    from django.conf import settings
        ##    apps.populate(settings.INSTALLED_APPS)

        with ReloadEmbeddedUrlConf(urlConf):
            pass

    def __registry_delete__(self):
        self.server_shutdown()

        from phsite.network.embedded import uninstall
        uninstall()

    def getStatus(self):
        return '{%s: %s}\n%s' % (self.__class__.__name__, self.runningState(),
                                 GetURLConf(indent = '  '))

@contextmanager
def ReloadEmbeddedUrlConf(urlConf = None):
    try: from django.core.urlresolvers import clear_url_caches
    except ImportError:
        from django.urls.base import clear_url_caches

    clear_url_caches()

    # Hack in the settings file, because this service doesn't compartmentalize
    # the project right in embedded mode.
    #template_dir = system.modulePath(system.lookup(__name__))
    #template_dir = template_dir.folder.folder.templates
    #template_dir = str(template_dir)

    from os.path import dirname, join as joinpath
    template_dir = dirname(dirname(__file__))
    template_dir = joinpath(template_dir, 'templates')

    try:
        if template_dir not in settings.TEMPLATE_DIRS:
            settings.TEMPLATE_DIRS += (template_dir,)

    except AttributeError:
        pass


    # XXX :skip: obseleted reference:
    # if 'web.application' not in settings.INSTALLED_APPS:
    #     settings.INSTALLED_APPS += ('web.application',)


    # Note: this has become necessary because the settings is abstracted
    # into the application configuration.
    if urlConf is not None:
        settings.ROOT_URLCONF = urlConf


    try: yield
    finally:
        # from web.stuph.embedded import urls
        #from web.stuph import urls
        #reload(urls)

        # todo: put them back into global_settings.ROOT_URLCONF,
        # wrap them in a hashable entity so that the object caching
        # internals work properly (urls.urlpatterns is a list).
        #from django.core import urlresolvers
        # First, dereference the previous '^/'-wrapped resolver object from the memory
        #settings.ROOT_URLCONF = urlresolvers.RegexURLResolver('', urls.urlpatterns)
        #settings.ROOT_URLCONF = tuple(urls.urlpatterns) # tuple(...)?

        pass


def GetURLConf(indent = ''):
    from django.core.urlresolvers import get_resolver, get_urlconf, RegexURLResolver, RegexURLPattern
    def formatResolver(r, indent):
        result = ['[module:%s (%s:%s)] %s' % (r.urlconf_name, r.app_name,
                                     r.namespace, r.regex.pattern)]

        indent2 = indent + '  '
        for url in r.url_patterns:
            if isinstance(url, RegexURLResolver):
                result.append(formatResolver(url, indent2))
            elif isinstance(url, RegexURLPattern):
                if url.name:
                    result.append('%s: %s' % (url.name, url.regex.pattern))
                else:
                    result.append(url.regex.pattern)

        tab = '\n' + indent
        return tab.join(result)

    return formatResolver(get_resolver(get_urlconf()), indent)

def importModule(name):
    return __import__(name, locals(), globals(), [''])


def djangoConfig():
    ValidactivateOnce()

    from django.conf import settings
    # Doesn't affect the construction of the middleware.
    # Q: What about getting the config before activation, and setting then?
    # settings.DEBUG = debug

    # Moved out of constructor because handler relies on it being configured sooner.
    from django.apps import apps
    apps.populate(settings.INSTALLED_APPS)


# Install Front End.
DjangoService.manage()


# Internal WSGI requests (for internal redirects).
class InternalServer(BaseServer):
    @classmethod
    def doRequest(self, serverAddress, method, path, input, **headers):
        server = self(serverAddress, self.RequestHandler)
        buf = BytesIO()

        server.set_request(buf, method, path, input, **headers)
        server._handle_request_noblock()

        return buf

    @classmethod
    def doRedirectRequest(self, request, path):
        qs = getQueryString(request.GET)
        if qs:
            path += '?'
            path += qs

        input = request.raw_post_data

        return self.doRequest(request.META['SERVER_ADDRESS'],
                              request.method, path, input,
                              **request.headers)


    def set_request(self, wfile, method, path, input, **headers):
        self.request = [wfile, method, path, input, headers]

    def get_request(self):
        return self.request


    class RequestHandler(WSGIRequestHandler):
        def setup(self):
            (wfile, method, path, input, headers) = self.request

            # Reconstruct the HTTP request message.
            rfile = BytesIO('%s %s HTTP/1.1\r\n' % (method, path))

            for (key, value) in headers.items():
                rfile.write('%s: %s\r\n' % (key, value))

            rfile.write('\r\n')

            if input:
                rfile.write(input)

            self.rfile = rfile
            self.wfile = wfile


def getQueryString(o):
    return '&'.join('%s=%s' % kv for kv in o.items())


requestInternal = InternalServer.doRequest
redirectInternal = InternalServer.doRedirectRequest


class httpBufferResponse(baseHTTPResponse):
    '''
    We can't use HTTPResponse directly because its constructor
    expects a socket-like object.

    '''

    def __init__(self, fp, *args, **kwd):
        # This is a hack to get around the sock-only constructor.
        baseHTTPResponse.__init__(self, self._bufferWrapper(fp),
                                  *args, **kwd)

    class _bufferWrapper:
        def __init__(self, fp):
            # self.makefile = (lambda *args, **kwd: fp)
            self.fp = fp
        def makefile(self, *args, **kwd):
            return self.fp


class requestInternalNative:
    '''
    menu:
        context($trigger)::
            r = request$internal(request, '/page/gen/site/menu')

            response['status'] = r.status
            response['content'] = r.content

            '''

    def __new__(self, frame, request, path):
        from stuphos.structure import RequestAdapter
        if not isinstance(request, RequestAdapter):
            raise TypeError(f'argument 1: {type(request).__name__}')

        task = frame.task

        @self._pool.task(task)
        def request(_):
            response = redirectInternal(request._request, path)

            with task.billableComputation():
                return self._response._build(request, response,
                                             method = request._request.method)


    class _response(writeprotected):
        @classmethod
        def _build(self, request, response, method = None):
            response = httpBufferResponse(response, method = method)
            return self(request, response.status, response.reason,
                        response.headers, response.read())

        def __init__(self, request, status, message, headers, content):
            self._request = request
            self._status = status
            self._message = message
            self._headers = headers
            self._content = content

        def __repr__(self):
            return '<response %d %r>' % (self.status, self.message)


        @property
        def request(self):
            return self._request

        @property
        def status(self):
            return self._status
        code = status

        @property
        def message(self):
            return self._message
        reason = message

        @property
        def headers(self):
            from stuphos.kernel import vmNewMapping
            return vmNewMapping(*self._headers.items())

        @property
        def content(self):
            return self._content

_requestInternalNative = requestInternalNative



# Standard implementation of Command Manager.
__all__ = ['StandardCommandManager']

from stuphmud import server as mud

from stuphmud.server.player.interfaces import CommandCenter
from stuphos.etc.tools import getLineFromCache, functionLines, functionName
import imp

SYSLOG_FILE = '../log/syslog'
def readSyslog():
    from os import getenv
    file = getenv('STUPH_LOGFILE', SYSLOG_FILE)
    return filterSyslog(open(file).read())

def filterSyslog(text):
    return text.replace('\x00', '').replace('\n', '\r\n')

##  try: from mud.tools import functionLines, functionName
##  except ImportError:
##      def notimpl(*args, **kwd):
##          raise NotImplementedError
##
##      functionLines = functionName = notimpl

def listCommands(commands):
    # Generates two-column lines.
    commandMap = dict()
    for v in list(commands.values()):
        commandMap[v] = list()

    # Collect names for each function -- Inversion.
    for (k, v) in commands.items():
        commandMap[v].append(k)

    # Go through each function, generating a listing for all the names and descr of func.
    for (func, names) in commandMap.items():
        func = func.current # registration
        try: func = func.command
        except AttributeError:
            pass

        doc = func.__doc__ or ''
        if doc:
            doc = doc.replace('\n', '')

        # yield functionName(func), doc
        yield func.__name__, doc

        # Zip side-by-sides.
        names.sort() # Alphabetically
        desc = list(map(str, functionLines(func, max = len(names), trim_decorators = True)))

        for (a, b) in sideBySide(names, desc):
            yield '    ' + a, b

    # Generate empty line.
    yield ('', '')

def sideBySide(a, b):
    '(unzip list-first-column list-second-column)'

    la = len(a)
    lb = len(b)

    for x in range(la):
        # Python 2.6 syntax capability:
        # i = a[x] if x < la else ''
        # j = b[x] if x < lb else ''
        if x < la:
            i = a[x]
        else:
            i = ''

        if x < lb:
            j = b[x]
        else:
            j = ''

        yield i, j

class StandardCommands(CommandCenter):
    'Some sample standard commands provided here.'
    # Only compatible with a CmdDict multiple-inheritance.

    # @MinimumLevel(LVL_IMPL)
    # @Qualifier(lambda peer, *args:peer.avatar and peer.avatar.implementor)
    @staticmethod
    def doPageSyslog(peer, cmd, argstr):
        'Page the ENTIRE SYSLOG!!'

        if peer.avatar and peer.avatar.implementor:
            peer.page(readSyslog())
            return True

    @staticmethod
    def doShowBridgedEvents(peer, cmd, argstr):
        'Display aspects on bridged mud events.'

        if peer.avatar and peer.avatar.implementor:
            try: from stuphos.system.api import game
            except ImportError:
                from stuphos.system import game

            from stuphmud.server import bridge

            bridge.showBridgedEvents(game.bridgeModule(), page = peer.page, endline = '\r\n')
            return True

    @staticmethod
    def doShowHeartbeatTasks(peer, cmd, argstr):
        'Show heartbeat tasks.'

        if peer.avatar and peer.avatar.implementor:
            from stuphmud.server import tasks
            tasks.show(peer.page)
            return True

    @staticmethod
    def doReloadInterpreter(peer, cmd, argstr):
        'Reload the player module, all the interpretation, and re-interpret player.'

        if peer.avatar and peer.avatar.implementor:
            from stuphmud.server import player
            imp.reload(player).interpret(peer)
            return True

    def doShowCommands(self, peer, cmd, argstr):
        'Displays command listing of this object.'

        # Generate lines to be printed side-by-side with names.
        width = 30, 80
        fmt   = '%%-%(left)d.%(left)ds %%-%(right)d.%(right)ds' % {'left':width[0], 'right':width[1]}

        peer.page('\r\n'.join(fmt % line for line in listCommands(self.c)) + '\r\n')
        return True

    def doCassiusConnection(self, peer, cmd, argstr):
        if peer.avatar and peer.avatar.supreme:
            cas = self.openCassius()

            # Python 2.6 syntax capability:
            # argstr = argstr.strip() if argstr else ''
            argstr = argstr and argstr.strip() or ''

            if argstr == 'close':
                # Delete installation known to mud, and our cas builtin.
                del builtin.cassius_connection, builtin.cas

                print('&D >&N', cas, 'Deleted!', file=peer)
                return True

            if argstr == 'reopen':
                del builtin.cassius_connection, builtin.cas

                # Re-open and fall thru...
                # reload(telnet)
                cas = self.openCassius()

            elif argstr == 'bind':
                cas.adapters.add(telnet.bind(peer))

                print('Okay.', file=peer)
                return True

            elif argstr == 'unbind':
                print('No idea how to do that!', file=peer)
                return True

            elif argstr:
                # Send to connection.
                cas.sendline(argstr)

                print('&D >&N', argstr, file=peer)
                return True

            print(cas, file=peer)
            return True

        # Not supreme, not allowed!

    def doMainMenu(self, peer, cmd, argstr):
        mainMenu = mud.player.namespace.getGlobals().get('Menu')
        if not mainMenu:
            from person.menu import Main as mainMenu

        return True

    # These have just gone obselete.
    # def setup(self):
    #     # CommandSetup.setup(self)
    #     a = self.insertOverridingAll

    #     a('page-*syslog',          self.doPageSyslog        )
    #     a('show-br*idged-events',  self.doShowBridgedEvents )
    #     a('show-t*asks',           self.doShowHeartbeatTasks)
    #     a('reload-*interpreter',   self.doReloadInterpreter )
    #     a('interpret*',            self.doReloadInterpreter )
    #     a('show-comm*ands',        self.doShowCommands      )
    #     a('main*-menu',            self.doMainMenu          )

# Export this:
##  class StandardCommandManager(CommandManager):
##      'Implements a command manager (abbreviation-managed command dictionary) with standard commands.'
##
##      def setup(self):
##          StandardCommandMixin.setup(self)
##
##          try:
##              # self.importCommands('commands')
##              self.importCommands('mud.player.interfaces.commands')
##
##          except ImportError:
##              pass
##
##      def __init__(self, *args, **kwd):
##          CommandManager.__init__(self, *args, **kwd)



# Deployment Telnet Dispatch Module
from stuphmud.server.player.db.managedfs import getCheckedPlayerFilename, PERSONAL_PLRFILE_DIR
from stuphmud.server.player.telnet import DispatchModuleBase

deploymentApi = runtime.MUD.Deployment.API
class LiveDeployDispatchModule(DispatchModuleBase):
    MULTITHREAD = True

    class PlayerFileArchive(deploymentApi.getArchiveBaseClass()):
        def __init__(self, peer, archive_name):
            plrfile = getCheckedPlayerFilename(peer, archive_name, PERSONAL_PLRFILE_DIR)
            FileArchive.__init__(self, plrfile, archive_name)

    class WizardProgress(deploymentApi.getProgressStreamBaseClass()):
        EOL = '\r\n'

        def __init__(self, player, synchronize):
            ProgressStream.__init__(player, synchronize)
            self.player = player

        def getStream(self):
            # This way, the player can be updated even if the connection re-logs in!
            return self.player.peer

    def _doDeployPersonalArchive(self, peer, digest, config_name, archive_name, update_message = None):
        config = deploymentApi.LoadDeploymentConfig(config_name)

        # todo: lock on config -- to prevent overlapping deploy conflicts
        if self._validateMessage(digest, config.key, peer.avatar.name, config_name, archive_name):
            deploymentApi.LiveDeploymentUrlFromConfigAndArchiveOnMultithread \
                (self.MULTITHREAD, config, self.PlayerFileArchive(peer, archive_name),
                 updateMessage = update_message)
                # progress_proc = self.WizardProgress(peer.player, self.MULTITHREAD)

    _doDeploypersonalarchive = _doDeployPersonalArchive
    _doDeploy = _doDeployPersonalArchive



# Installation of Computation Core.

from stuphos.xmlrpc import getHost
from stuphos.xmlrpc import host

from stuphos.xmlrpc.config import NotConfigured


from stuphos import getConfig
from stuphos.etc import isYesValue

if isYesValue(getConfig('metal', 'VM')):
	from ph.emulation.metal import Machine
	from ph.emulation.metal import MachineCore
	from ph.emulation.metal import Native
	from ph.emulation.metal import TaskControl

	from ph.emulation.metal import Heartbeat

	from ph.emulation.metal import Processor
	from ph.emulation.metal import Subroutine

else:
	from ph.emulation.machine.virtual import Machine
	from ph.emulation.machine.virtual import MachineCore
	from ph.emulation.machine.virtual import Native
	from ph.emulation.machine.virtual import Pure
	from ph.emulation.machine.virtual import TaskControl
	# from ph.emulation.machine.virtual import startTask

	from ph.emulation.machine.virtual import machine
	from ph.emulation.machine.virtual import checkActiveTasks
	from ph.emulation.machine.virtual import checkActiveTimers
	from ph.emulation.machine.virtual import BypassReturn
	from ph.emulation.machine.virtual import OuterFrame
	from ph.emulation.machine.virtual import Continuation
	from ph.emulation.machine.virtual import resolve_procedure
	from ph.emulation.machine.virtual import EmptyStackError

	from ph.emulation.machine import currentVMTask
	from ph.emulation.machine import vmCurrentTask
	from ph.emulation.machine import vmCheckAccess
	from ph.emulation.machine import vmNewMapping
	from ph.emulation.machine import vmNewSequence
	from ph.emulation.machine import nativeObject
	from ph.emulation.machine import nativeMemoryObject
	from ph.emulation.machine import deletable
	from ph.emulation.machine import baseStringClass
	from ph.emulation.machine import StringValue
	from ph.emulation.machine import InlineTextClass
	from ph.emulation.machine import TaskCreation

	from ph.emulation.machine.memory import AutoMemoryObject
	from ph.emulation.machine.memory import AutoMemoryMapping
	from ph.emulation.machine.memory import AutoMemorySequence
	from ph.emulation.machine.memory import AutoMemoryNamespace
	from ph.emulation.machine.memory import protectedMemoryCopy
	from ph.emulation.machine.memory import protectedMemoryLoad
	from ph.emulation.machine.memory import MemoryMapping
	from ph.emulation.machine.memory import MemorySequence


	from ph.emulation.machine.heartbeat import Heartbeat

	from ph.emulation.operation.application import Processor
	from ph.emulation.operation.application import Subroutine
	from ph.emulation.operation.application import Generator
	from ph.emulation.operation.application import FastLocals


	try: from ph.emulation.machine.parallel import pool
	except ImportError: pass


from ph.emulation.security import RelationNetwork


# debugOn()
try:
	from ph.interpreter import mental as interpreter

	# from stuphos.runtime import Undefined
	from ph.interpreter.mental import Undefined

	from ph.interpreter.mental import Assembly
	from ph.interpreter.mental import CallMethodError
	from ph.interpreter.mental import Girl
	from ph.interpreter.mental import Ella # New symbology.
	from ph.interpreter.mental import Agent
	from ph.interpreter.mental import ParallelTask
	from ph.interpreter.mental import Programmer
	from ph.interpreter.mental import Script
	from ph.interpreter.mental import Volume
	from ph.interpreter.mental import callGirlMethod
	from ph.interpreter.mental import executeGirl
	from ph.interpreter.mental import getLibraryCore
	from ph.interpreter.mental import grammar
	from ph.interpreter.mental import invokeLibraryMethod
	from ph.interpreter.mental import newModuleTask
	from ph.interpreter.mental import callStringValueMethod

	from ph.interpreter.mental import findUser
	from ph.interpreter.mental import findUserByName
	from ph.interpreter.mental import findCurrentUser

	from ph.interpreter.mental import native
	from ph.interpreter.mental import nullproc

	from ph.interpreter.mental.objects import Library
	from ph.interpreter.mental.objects import LibraryNode
	from ph.interpreter.mental.objects import Instance
	from ph.interpreter.mental.objects import GirlSystemModule
	from ph.interpreter.mental.objects import SyntheticClass
	from ph.interpreter.mental.objects import constrainStructureMemory
	from ph.interpreter.mental.objects import NoSymbolError

	from ph.interpreter.mental.grammar import GrammaticalError

	from ph.interpreter.mental.compiler import WMBasedCompiler

	from ph.interpreter.mental.native import sleep, create, call
	from ph.interpreter.mental.native import _securityContext as securityContext
	from ph.interpreter.mental.native import _accessItemsNative as accessItemsNative
	from ph.interpreter.mental.native import _setTimeout

	from ph.interpreter.mental.ipc import cleanupObject

	from ph.interpreter.mental.library.model import GirlCore
	from ph.interpreter.mental.library.model import LIB_ROOT
	from ph.interpreter.mental.library.extensions import serverChannel

	def getLibraryViews():
		# The views module relies on django settings which might not
		# be loaded when this kernel module is loaded.
		from ph.interpreter.mental.library import views
		return views

	def wm(*args, **kwd):
		global wm
		from ph.interpreter.mental.library.views import wm
		return wm(*args, **kwd)

	# from ph.interpreter.mental.library.views import LibraryView

# Disable interpreter.
except ImportError as e:
	# todo: print stack trace
	print(f'[stuphos.kernel:interpreter] {e}')
	# debugOn()


##    XXX MOBILE_TRIGGER_TYPES :skip:
##    XXX getMobileTriggerSheet :skip:
##    XXX getMobileTriggerType :skip:
##    XXX libraryViews :skip:


from stuphos.runtime.architecture.routines import importAllInit
importAllInit(configuration.VM.kernel, '[VM:kernel]', globals())



from stuphlib.dblib import SingleRecordReader
from stuphlib.binary import Record, BuildRecordClass

from datetime import datetime

NUM_CLAN_RANKS = 10

class ClanReader(SingleRecordReader):
    def parseRecord(self):
        # Read one record per file -- #<vnum> already read.
        self.recordField('alias', self.tildeString())
        self.recordField('name', self.tildeString())
        self.recordField('colorized_name', self.tildeString())

        (donation_room, entrance_room, recall_room,
         range_low_room, range_high_room) = self.readLine().split()

        self.recordField('donation_room'  , donation_room   )
        self.recordField('entrance_room'  , entrance_room   )
        self.recordField('recall_room'    , recall_room     )
        self.recordField('range_low_room' , range_low_room  )
        self.recordField('range_high_room', range_high_room )

        # I believe these are currently unused.
        (store_object, last_updated, updated_by) = self.readLine().split()

        self.recordField('store_object', store_object )
        self.recordField('last_update' , last_updated )
        self.recordField('updated_by'  , updated_by   )

        # Read rank titles.
        ranks = []
        for x in range(NUM_CLAN_RANKS):
            male = self.tildeString()
            female = self.tildeString()

            rank = dict(male = male, female = female)

            self.recordField('rank%d' % (x + 1), rank)
            ranks.append(rank)

        self.recordField('ranks', ranks)

ClanPowers = [('apply'  , -1),
              ('ban'    ,  8),
              ('demote' ,  8),
              ('deposit',  2),
              ('donate' ,  0),
              ('enlist' ,  5),
              ('expel'  ,  8),
              ('help'   , -1),
              ('info'   , -1),
              ('log'    ,  0),
              ('promote',  6),
              ('ranks'  ,  0),
              ('recall' ,  0),
              ('resign' ,  6),
              ('roster' ,  8),
              ('set'    ,  9),
              ('title'  ,  0),
              ('unban'  ,  8),
              ('who'    ,  0),
              ('withdraw', 6)]

class ClanBanRec(Record, metaclass=BuildRecordClass):
    Fields = [('idnum', Record.Long  ), # Player banned
              ('by'   , Record.Long  ), # Who did the banning
              ('when' , Record.TimeT )]

    BANTIME_FORMAT = '%a, %b %e (%T)'

    def __str__(self):
        timestr = datetime.fromtimestamp(self.when).strftime(BANTIME_FORMAT)
        return '&C%-20.20s &Nby &C%-20.20s&N on %s' % \
               (self.idnum, self.by, timestr)

def ReadClanBankFile(fp):
    for line in fp:
        if line == '':
            break

        line = line.strip()
        if line == '$':
            break

        (vnum, amount) = line.split()
        yield (int(vnum),
               int(amount))

def ReadClanInformation(fp):
    return fp.read()

# Indexing.
from stuphlib.dblib import LocalIndex, INDEX_FILE
from stuphlib.wldlib import Record as WorldRecord, Initializer
from stuphlib.binary import ReadFile

ClanDataPrefix = 'data'
ClanDataSuffix = 'dat'

ClanInfoPrefix = 'info'
ClanInfoSuffix = 'inf'

ClanBanPrefix  = 'ban'
ClanBanSuffix  = 'ban'

CLAN_BANK_FILE = 'bank'

class ClanIndex:
    def clanFileName(self, prefix, vnum, suffix):
        return self.joinpath(prefix, '%d.%s' % (vnum, suffix))

    def clandatafile(self, vnum):
        return self.clanFileName(ClanDataPrefix, vnum, ClanDataSuffix)
    def claninfofile(self, vnum):
        return self.clanFileName(ClanInfoPrefix, vnum, ClanInfoSuffix)
    def clanbanfile(self, vnum):
        return self.clanFileName(ClanBanPrefix , vnum, ClanBanSuffix)

    def openClanDataFile(self, vnum):
        return self.open(self.clandatafile(vnum))
    def openClanInfoFile(self, vnum):
        return self.open(self.claninfofile(vnum))
    def openClanBanFile(self, vnum):
        return self.open(self.clanbanfile(vnum))

    def clanbankfile(self):
        return CLAN_BANK_FILE
    def openClanBankFile(self):
        return self.open(self.clanbankfile())

class ClanLocalIndex(ClanIndex, LocalIndex):
    pass

class ClanBank:
    def __init__(self, balance):
        self.balance = balance
    def __repr__(self):
        return '%s: %d gold' % (self.__class__.__name__,
                                self.balance)

class Clan(WorldRecord):
    class Meta(WorldRecord.Meta):
        __attributes__ = ('name', 'bank', 'ranks', 'information')

    def __init__(self, base, vnum):
        Initializer.__init__(self)
        self.base = base

        (self) << (vnum)

    # Implementation of Record.
    ReaderClass = ClanReader

    def loadFromReader(self, reader):
        reader.parseFile()
        recordSet = reader.consumeRecordSet()
        self.vnum = list(recordSet.keys())[0]
        self.__dict__.update(recordSet[self.vnum])

    def loadFrom(self, vnum):
        self.loadFromFile(self.base.openClanDataFile(vnum))
        self.information = ReadClanInformation(self.base.openClanInfoFile(vnum))
        self.ban_list = list(ReadFile(self.base.openClanBanFile(vnum),
                                      ClanBanRec))

def LoadClans(clan_base, index_file):
    index = ClanLocalIndex(clan_base)

    clans = dict()
    for (vnum, fix) in index.iterindex(index_file):
        clans[vnum] = Clan(index, vnum)

    # Load bank data.
    for (vnum, bank) in ReadClanBankFile(index.openClanBankFile()):
        clans[vnum].bank = ClanBank(bank)

    return clans

def main(argv = None):
    from optparse import OptionParser
    from code import InteractiveConsole as IC

    parser = OptionParser()
    parser.add_option('-i', '--index-file', default = INDEX_FILE)
    parser.add_option('-b', '--clan-base')
    (options, args) = parser.parse_args(argv)

    index_file = options.index_file
    clan_base = options.clan_base
    assert clan_base

    clans = LoadClans(clan_base, index_file)

    import readline
    IC(locals = dict(clans = clans)).interact(banner = '')

##        print '========================================================================'
##        print 'Currently Banned:'
##        print '========================================================================'
##                print '&Y%3d. %s' % (x, ban)

if __name__ == '__main__':
    main()



# Remote Execution Telnet Module
from stuphmud.server.player.telnet import DispatchModuleBase
from stuphmud.server.player.telnet.dispatch import LoadKeyfile

from stuphmud.server.player import getPlayerScope, playerAlert
from stuphmud.server.player.db.managedfs import getCheckedPlayerFilename, KEYFILE_DIR, PERSONAL_PLRFILE_DIR

import os.path
from new import module as NewModule

class RExecDispatchModule(DispatchModuleBase):
    def validatePlayerCommandMessage(self, peer, digest, message):
        # XXX Should warn when the digest doesn't match?
        return self._validateMessage(digest, self.getPlayerKey(peer), peer.avatar.name, message)

    def getPlayerKey(self, peer):
        # Retrieve player key from personal file store.
        filename = getCheckedPlayerFilename(peer, 'rexec.key', KEYFILE_DIR)
        return LoadKeyfile(filename)

    def getPlayerId(self, peer):
        return '%s%s' % (peer.avatar.name, peer.avatar.idnum)

    def getScopeModule(self, peer):
        scope = getPlayerScope(peer)
        try: return scope.rexec
        except AttributeError:
            rexec = NewModule('%s.namespace' % __name__)
            scope.rexec = rexec
            return rexec

    def getPlayerScope(self, module, peer):
        try: scope = module.playerScope
        except AttributeError:
            scope = module.playerScope = {}

        playerId = self.getPlayerId(peer)
        try: return scope[playerId]
        except KeyError:
            playerMod = NewModule('%s.namespace.%s' % (__name__, playerId))
            scope[playerId] = playerMod
            return playerMod

    def getScope(self, peer):
        module = self.getScopeModule(peer)
        immediate = self.getPlayerScope(module, peer)

        # Install environment.
        immediate.this = peer
        immediate.me = peer.avatar

        return (module.__dict__, immediate.__dict__)

    def _doExecute(self, peer, digest, sourceCode):
        if self.validatePlayerCommandMessage(peer, digest, sourceCode):
            # Evaluator Shell?
            code = compile(sourceCode, '<rexec>', 'single')
            (globals, locals) = self.getScope(peer)

            def executeCode(peer):
                exec(code, globals, locals)

            peer.interpreter.withPeerHead(peer, executeCode)

    def _doExecuteFile(self, peer, digest, filename):
        if self.validatePlayerCommandMessage(peer, digest, filename):
            plrfile = getCheckedPlayerFilename(peer, filename, PERSONAL_PLRFILE_DIR)
            sourceCode = open(plrfile).read()
            sourceCode = sourceCode.replace('\r', '')
            code = compile(sourceCode, filename, 'exec')

            (globals, locals) = self.getScope(peer)
            globals['__file__'] = plrfile # filename

            def executeCode(peer):
                exec(code, globals, locals)

            peer.interpreter.withPeerHead(peer, executeCode)

    _doExecutefile = _doExecuteFile



class HistoryManager:
	## All history subroutines work on the peer history buffer so require the peer be passed as argument.
	def historyObject(self, peer, create = True):
		# debugOn()
		history = getattr(peer, 'history', None)
		if not history:
			assert create, AttributeError('history')
			history = peer.history = []

		return history

	def historyOverride(self, peer, cmd, argstr):
		'''
		Implements a variety of history-handling functions, executing inline with shellCommand
		by modifying its arguments prior to interpreting them, so as to avoid recursion.

		Returns new parsed command elements or those given.

		These commands are currently parsed only by the shell interpreter and apply only to
		commands invoked therein.  Commands overridden do not effect the outside calling mud
		context.

		'''

		if cmd != '!':
			# Save to history.
			self.historyAdd(peer, cmd, argstr)
			return cmd, argstr

		# args is stripped form of argstr
		args = argstr.strip()

		if not args:
			# Maybe execute last command history -- Or pass to underlying mud code.
			return self.historyLast(peer) or (cmd, args)

		if args[0] == '!':
			# Show history given arguments.
			self.historyShow(peer, args[1:])
			return (None, None)

		elif args.isdigit():
			# Search for command in history and return.
			return self.historyLast(peer, int(args))

		# Todo: allow set by name, and lookup by name, history management.
		i = args.find('=')
		if i >= 0:
			name = args[:i].strip()
			args = args[i+1:].strip()

			if name.isdigit():
				# Replace history position.
				pos = int(name)

				# Make sure position is appropriate within range.
				history = self.historyObject(peer)
				n = len(history)

				if pos < 0 or pos > n:
					print('Position %d out of range! (%d max)' % (pos, n))
				else:
					history[-pos] = self.parseCommand(args)
					print('History %d replaced with %r.' % (pos, args), file=peer)

			elif name:
				# Player's variables
				variables = getattr(peer, 'variables', None)
				if variables is None:
					variables = peer.variables = dict()

				print('Replacing peer variables %s with %s' % (name, args), file=peer)
				variables[name] = args

			else:
				print('Unknown form: = %s' % (args), file=peer)

			return (None, None)

		# Not recognized.
		return cmd, argstr

	def historyLast(self, peer, n = 1):
		if type(n) is int or type(n) is str and n.isdigit():
			return self.historyObject(peer)[-int(n)]

	def historyAdd(self, peer, cmd, argstr):
		history = self.history = self.historyObject(peer)
		history.append((str(cmd or ''), str(argstr or '')))

	def historyShow(self, peer, args, history = None):
		if history is None:
			history = self.historyObject(peer)

		if history:
			end = len(history)

			def historyIterate():
				for x in range(end): # + 1?
					yield x, history[x]

			def joinParse(cmd, argstr):
				if cmd in '0123456789`~!@#$%^&*()-_=+{[}]:;"\'<,>.?/|\\':
					return cmd + argstr

				# Put a space in between.
				return '%s %s' % (cmd, argstr)

			peer.page('\r\n'.join('%-3d : %.80s' % (end - x, joinParse(*h)) for (x, h) in historyIterate()) + '\r\n')

	def historyFilter(self, peer, name = None, onlyArguments = False):
		history = self.historyObject(peer)
		if name is None:
			for (cmd, args) in history:
				yield onlyArguments and args or (cmd, args)

		elif type(name) is str:
			for (cmd, args) in history:
				if cmd == name:
					yield onlyArguments and args or (cmd, args)

		elif type(name) in (tuple, list):
			for (cmd, args) in history:
				if cmd in name:
					yield onlyArguments and args or (cmd, args)

	def historySubstitution(self, peer, line):
		'Eventually parse $<int> $name and $(complex.name) and ${expression block} lines to include the variables/history numbers.'

		def parseHistory(n):
			# Join the command and argument strings.
			return ''.join(self.historyLast(peer, n))

		return self.substitute(line, lookup = parseHistory)



#!/usr/local/bin/python
'Implements command searching with abbreviation management.  See CmdDict.'

# Override UserDict?
class CmdDict:
	'''
	Briefly, it encapsulates a command set dictionary that any command or
	abbreviation of a command will map to value.

	Insert control is provided for expressing abbreviated forms and those
	commands already present in the command set acceptable for overriding.

	And override is where one command abbreviation mapping is replaced with
	another according to the rules of insert[Overriding].

	'''

	# What method to use when using decorator assignment of functions.
	assignmentMethod = 'OverridingAll' # 'Overriding' # ''

	def __init__(self, c = None, assignment = None):
		# Forcing dict type:
		if c is None:
			c = self.c = dict()
		else:
			self.c = c

		self.__getitem__ = c.__getitem__
		self.__setitem__ = c.__setitem__

		# Decide assignment method.
		self.insertAssignment = getattr(self, 'insert' + (assignment or self.assignmentMethod))

	@staticmethod
	def abbreviations(*args):
		'''
		For each positional argument, generate a number of abbreviated
		forms, from longest to shorted, until the end 

		Each argument be either:
			- A 2-tuple:
			  ('command-name', end)
			  Where end is the ending index of the command name to stop abbreviations.

			- A string describing verb abbreviations (from lambda moo):
			  verb*name
			  The abbreviations are generated from the full name down to the asterisk.

			- A plain string.  All possible abbreviations will be generated down to
			  the first letter.  Useful for overriding the movement or violence commands.

		'''

		for m in args:
			if type(m) is tuple and len(m) == 2:
				# Only yield up to explicit column.
				m, end = m
				while len(m) >= end:
					yield m
					m = m[:-1]
			elif '*' in m:
				# Detect end column from 'verb*name' form.
				end = m.index('*') # -1 ?
				m = m.replace('*', '')
				while len(m) >= end:
					yield m
					m = m[:-1]
			else:
				# Use every possible abbreviation.
				while m:
					yield m
					m = m[:-1]

	class Registration(list):
		def __init__(self, cmddict, name):
			self.name = name
		def __repr__(self):
			return '<%s %r (%d assignments)>' % (self.__class__.__name__, self.name, len(self))

		@property
		def current(self):
			return self[-1]

		def __hash__(self):
			h = 0
			for r in self:
				h ^= hash(r)
			return h

		def __add__(self, assignment):
			self.append(assignment)
			return len(self)-1

		def __call__(self, *args, **kwd):
			return self.current(*args, **kwd) # overriding all

		def remove(self, index = None):
			if index is None:
				del self[:]
			else:
				try: list.__delitem__(self, index)
				except IndexError:
					pass

			return len(self) == 0

	def register(self, name, assignment):
		try: c = self.c[name]
		except KeyError:
			self.c[name] = c = self.Registration(self, name)

		return c + assignment

	def insert(self, m, a = None):
		'Insert abbreviations of command name into command set not overriding anything.'

		c = self.c
		a = a or m

		for x in self.abbreviations(m.lower()):
			# Overwrite not desired for any existing entries.
			if x not in c:
				c[x] = a

	def insertOverriding(self, m, a = None, *overriding):
		'''
		Insert only if command is either not already in command-set, or is part of overriding.

		Overriding arguments can be delivered as an abbreviation list:
		insertOverriding('the-*first-command', doCmd, *abbreviations('the-*second-command', 'the-*third-command))

		'''

		c = self.c
		a = a or m

		if overriding:
			s = set(self.abbreviations(overriding))

			for x in self.abbreviations(m.lower()):
				# Override all forms for those specified in trailing arguments.
				if x in s or x not in c:
					c[x] = a
		else:
			for x in self.abbreviations(m.lower()):
				# Reverse order invoking this function no overwrite worry.
				c[x] = a

	def insertOverridingAll(self, m, a = None):
		'Insert abbreviations of command name into command set overriding any and all.'

		c = self.c
		a = a or m

		r = []
		for x in self.abbreviations(m.lower()):
			r.append((x, self.register(x, a)))

		return r

	def lookup(self, cmd, default = None):
		'Looks for command name in the command set dictionary or returns a default value.'

		return self.c.get(cmd, default)

	def assign(self, *verbNames, **req):
		# Decorator for assignment, using free cell variable.
		assign = self.insertAssignment

		def invokeInsert(function):
			r = []
			for form in verbNames:
				for e in assign(str(form), function, **req):
					r.append(e)

			return self.Assignment(self, function, verbNames, *r)
			return function

		return invokeInsert

	assignRemoveable = assign

	class Assignment(list):
		def __init__(self, cmddict, function, verbNames, *registrations):
			list.__init__(self, registrations)
			self.cmddict = cmddict
			self.function = function
			self.verbNames = verbNames

		def __call__(self, *args, **kwd):
			return self.function(*args, **kwd)

		def remove(self): # , index = None):
			for (a, r) in self:
				self.cmddict.remove(a, r)

	@staticmethod
	def parse(cmd):
		'''Parses a MUD command considering non-alphanumeric first characters.
		Returns 2-tuple (command-name, argstr)'''

		cmd = cmd.lstrip()
		if not cmd:
			return (None, None)

		x = cmd[0]
		if not x.isalpha():
			return x, cmd[1:]

		x = cmd.find(' ')
		if x >= 0:
			return cmd[:x].lower(), cmd[x+1:]

		return cmd.lower(), None

	def remove(self, abbrev, index = None):
		try: r = self.c[abbrev]
		except KeyError: pass

		if r.remove(index = index):
			del self.c[abbrev]

	# def remove(self, match):
	# 	c = self.c
	# 	for n in [n for (n, v) in c.iteritems() if v == match]: # v is match?
	# 		c[n].remove(c, n)

	def removeByLookup(self, name):
		cmdfunc = self.lookup(name)
		if cmdfunc is not None:
			self.remove(cmdfunc)

# Testing
if __name__ == '__main__':
	from sys import stdin
	from consoleInteractive import *
	from Timing import Timing

	from pprint import pprint as pp
	from pdb import run, runcall

	cmds = CmdDict()

	@Timing('Multiple Command Load : %(SECONDS)d.%(MICRO)06d')
	def load(n, i = cmds.insertOverriding):
		for x in n:
			i(x)

	@Timing('Command Search : %(SECONDS)d.%(MICRO)06d')
	def lookup(n):
		n, argstr = cmds.parse(n)
		return cmds.lookup(n), argstr

	# Initial load.
	loadedCmds = list(input_cmdln())
	loadedCmds.reverse()

	for x in range(10):
		cmds.c = {}
		load(loadedCmds)

	def matchAll(x):
		buf = __import__('cStringIO').StringIO()
		for (k, v) in cmds.data().items():
			if v.startswith(x):
				print('%-20s' % k, ':', v, file=buf)
		page(buf.getvalue())

	for n in getline():
		if shellCommand(n, globals()):
			continue

		print(lookup(n))



# Other binary-format files.
from stuphlib.binary import Record, BuildRecordClass, ReadOneFromFile

# etc/bank
class NumberOfAccounts(Record, metaclass=BuildRecordClass):
    Fields = [('number', Record.SizeT)]

class BankAccountBalance(Record, metaclass=BuildRecordClass):
    Fields = [('balance', Record.SizeT)]

def ReadBankFile(fp):
    numAccounts = ReadOneFromFile(fp, NumberOfAccounts).number
    for x in range(numAccounts):
        yield ReadOneFromFile(fp, BankAccountBalance).balance

# etc/shopgold
class NumberOfShops(Record, metaclass=BuildRecordClass):
    Fields = [('number', Record.UInt)]

class ShopBankAccount(Record, metaclass=BuildRecordClass):
    Fields = [('gold', Record.UInt),
              ('bank', Record.UInt)]

def ReadShopGold(fp):
    # This is supposed to match the number of shops internally.
    numShops = ReadOneFromFile(fp, NumberOfShops).number

    # So it shall, sequentially (but with no validation).
    for x in range(numShops):
        yield ReadOneFromFile(fp, ShopBankAccount)

# Inspection Front End
if __name__ == '__main__':
    from sys import argv
    from os.path import basename
    if len(argv) == 2:
        filename = argv[1]
        base = basename(filename)

        if base == 'bank':
            nr = 0
            for balance in ReadBankFile(open(filename)):
                nr += 1
                print('#%-5d : $%10d' % (nr, balance))

        elif base == 'shopgold':
            nr = 0
            for shopBankAcct in ReadShopGold(open(filename)):
                nr += 1
                print('#%-5d : $%10d  [ $%10d ]' % (nr, shopBankAcct.gold,
                                                    shopBankAcct.bank))



"""
Please note that processes compiled with -pg profiling may not use the exec*
family of system calls (which this program does to initiate the reboot).

Invoking execve will cause the process to abort and print this message to stderr:
Profiling timer expired

"""
__all__=['Copyover']
# bootIntoMiniMUD=True # On reboot, give the new process the -m cmdln switch

import stuphos
import stuphmud.server.player

# XXX dependencies:
from stuphos.system.api import game, world, syslog

from os import execl, close, getenv, unlink, environ, rename # putenv as setenv
from os import chdir, getcwd
from os.path import join as joinpath, dirname
from marshal import dump, load
from traceback import print_exc
from time import time as now
from pprint import pformat

try : from fcntl import fcntl, F_SETFD, F_GETFD, FD_CLOEXEC
except ImportError:pass

def setenv(n, x):
    'This puts the string value of object into the environment for given key.'
    # environ is an instance of the _Environ class, which wraps getenv/putenv
    # in index or slice methods.  So this calls putenv.  Why doesn't putenv
    # work when invoked directly!?
    environ[n]=str(x)

COPYOVERKEY = 'STUPH_COPYOVER_KEY'
STUPHBIN    = 'bin/stuph' # sys.argv[0]
EXENAME     = STUPHBIN # 'stuph'
DFLTKEY     = 'default'

def persistSocket(fd):
    'This is broken because ~1 returns -2!  Fortunately, sockets seem to be !FD_CLOEXEC initially.'
    oldbits=fcntl(fd, F_GETFD)
    newbits=oldbits & ~FD_CLOEXEC

    #print oldbits, '&', '~%d' % FD_CLOEXEC, '(%d)' % ~FD_CLOEXEC, '=', newbits

    ##fcntl(fd, F_SETFD, newbits)
    return fd

class Copyover:
    """
    Implementation of Network Session Persistance.

    Do a clean reboot (or copyover):
       Save file descriptors, (unset close on exec), write to file (etc/network-sessions.pkl)
       Say goodbye to everyone??
       Re-execute the mud in this process-space.

       Possibly make the copyover process part of the serialization/deserialization of this class.

    The instance of a Copyover class is constructed with unique identifier of the netsessions store,
    and the methods implemented can be used for either initiating a copyover reboot, or recovering
    from a copyover, and the sharing of the netsessions key across the process (the environment).
    """

    # If path does not exist, disable Copyover?
    NETSESSION_PATH='etc/network-sessions'
    DefaultKey=DFLTKEY

    def __init__(self, exepath, key=DefaultKey, *args):
        self.exepath = exepath
        self.exename = exepath
        self.key     = key
        self.store   = self.findStore(key)

        # Todo: From mud.config:
        # Force server:
        self.args    = ('-server', ) + args

        # Optional setting:
        if self.isMiniMUD():
            self.args += ('-m',)

        # Specifying the mudlib dir will cause changing into it on post.
        # self.args += ('-l', getcwd())

    def isMiniMUD(self):
        setting = globals().get('bootIntoMiniMUD')
        if setting is not None:
            return bool(setting)

        setting = stuphos.getConfig('mini-mud', 'Copyover')
        if setting is not None:
            setting = setting.lower()
            if setting in ('true', 'on', 'yes', '1'):
                return True
            if setting in ('false', 'off', 'no', '0'):
                return False

        # Is currently in minimud
        try: from stuphos.system.api import game
        except ImportError: mini_mud = game.mini_mud
        else:
            if mini_mud():
                return True

        return False

    ## Store management:
    def findStore(self, key):
        'Give a file or system name to the netsession key.'
        if key=='default': # or key.upper()=='DEFAULT'
            return self.NETSESSION_PATH+'.bin'

        return joinpath(self.NETSESSION_PATH, str(key)) # + '.bin'

    def dumpToStore(self, data):
        # dump(data, open(self.store, 'wt'))
        from stuphmud.server.zones.freeze import dumpFile
        dumpFile(data, self.store)

    def loadStore(self):
        # return load(open(self.store))
        from stuphmud.server.zones.freeze import loadFile
        return loadFile(self.store)

    def destroyStore(self):
        # unlink(self.store)
        rename(self.store, '%s.trash' % self.store)

    def getStore(self):
        try: return self._loaded_store
        except AttributeError:
            st = self._loaded_store = self.loadStore()
            return st

    ## Persistant data:
    def avatarData(self, a):
        if not a: # or a.peer.state!='Playing'
            return {}

        return {
            'name'   : a.name,
            'room'   : a.room is not None and a.room.vnum or -1,
        }

    def playerData(self, p):
        'Do a selection of data from the player for persisting.'
        if p.state!='Playing':
            return None

        data={
            'socket' : persistSocket(p.descriptor),
            'host'   : p.host,
            'state'  : p.state,
            'input'  : p.input_queue,
        }

        data.update(self.avatarData(p.avatar))

        return data

    def mudData(self, options):
        data = {
            # 'mother'  : -1, # world.getMother(),
            'players' : [_f for _f in map(self.playerData, world.player.users()) if _f],
            'time'    : now(),
        }

        if options.world or options.complete:
            from stuphmud.server.zones.freeze import dumpWorldData
            data['world'] = dumpWorldData(options)

        return data

    ## Reboot process:
    def initiate(self, options):
        'Save player data and invoke the reboot procedure (doReboot).'
        syslog('Initiating Copyover Reboot...')

        # Write persistant data to store
        data = self.mudData(options)
        #game.syslog('\n'+pformat(data))
        self.dumpToStore(data)

        # Save everyone to playerfile, giving it that special key.
        # This will eventually be separated with some kind of database interface.
        game.library.save_all_players('Closing Player File')

        # Do this step so that the mother-socket regenerates itself rather than being saved.
        # If it did persist across exec, would that mean the backlog of awaiting acceptees
        # would remain?  (who cares)
        #
        # What about platforms where the child connections require that the mother socket
        # stay open?  Is this a case on any system?
        #
        # It is important to code this correctly, because all of the players are already
        # closed (meaning that the descriptor_data don't exist anymore).
        game.shutdownMother()

        # save_mud_time

        # Initiate reboot process.  Perhaps notify log that we're really doing it now..
        # If we want to tell everyone we're initiating a copyover, we'll have to have
        # done that during the previous heartbeat pulse, because we're going down NOW!

        # Step out of this directory and into parent directory, because we'll want
        # to chdir right back into it.
        # The real reason
        chdir(dirname(getcwd()))
        # ...Or, since the chdir facility exists with -l option, just rely on that.
        # Seems more formal.

        setenv(COPYOVERKEY, self.key)
        execl(self.exepath, self.exename, *self.args)

    __call__ = initiate

    ## Recovery process:
    def recoverPlayer(self, p):
        # Instantiating a peer descriptor from this stratum requires an avatar to connect to:
        avatar = world.player.loadPlayer(p['name'], p['room'])
        peer   = world.player.emergePlayer(p['socket'], p['state'], avatar)

        # Check for deleted character??
        # The CON_GET_NAME login state also removes:
        #    PLR_WRITING | PLR_MAILING | PLR_CRYO | PLR_CREATING | PLR_INWAR
        #    AFF_GROUP
        # in case they were saved in the file.

        # unrent is Crash_load.  Additionally, the following steps pick up
        # where the enter-game code from nanny() leaves off.
        avatar.unrent()
        avatar.save()

        # Set peer.last_host to pc_store.host
        # Set peer.login to pc_store.last_logon

        # world.newConnection(peer)
        peer.host = p['host']
        mud.player.interpret(peer)

        for input in p.get('input') or []:
            peer.input = input

        # Notify log of each individual recovery..?  unrent/Crash_load notifies the log..
        print('&r*&yWelcome Back From Copyover&r*&N', file=peer)

        # XXX I think there's a problem with just calling this now...
        # Why doesn't enter game seem to install the programs??
        #
        # This is why I introduced playerActivation, which is removed from
        # any newConnection/enterGame nonsense, because these things won't
        # happen after copyover (yet the extension runtime still needs to
        # be awakened).
        if avatar.room:
            mud.player.enterGame(peer, avatar)

    def recoverMother(self, socket):
        """
        Instead of copying over the mother socket, it could be set Close-On-Exec.
        Since the stuphMUDGame (and hence FixupMother) isn't instantiated when the
        mud package module imports tools.reboot (hence initiating copyover recovery),
        we can't use fixupMother at this point.

        """
        # world.fixupMother(socket)

    def recover(self):
        # Load persistance data store (netsession)
        mudData = self.getStore()

        #game.syslog('\n'+pformat(mudData))
        syslog('Recovering from Copyover...')

        # Calculate time from mudData['time']

        # Fixup the mother socket
        # self.recoverMother(mudData['mother'])

        # Connect those players back into the mud, logging them in.
        # Bump off or re-authenticate those that don't jive.
        for p in mudData['players']:
            try    : self.recoverPlayer(p)
            except : print_exc()

        worldData = mudData.get('world')
        if worldData:
            from stuphmud.server.zones.freeze import loadWorldData
            loadWorldData(worldData)

        self.destroyStore()
        # gc.collect()

## Check for copyover when this module is loaded such that this module
## must be loaded after a copyover to ensure complete recovery.
#
## Issues of recovering from here:
##    Since this would occur during the mud/__init__.py script, the game hasn't
##    even been initialized yet: stuphMUDGame hasn't been instantiated, so the
##    mother socket isn't open, nothing's initialized, including the player file.
##
##    Perhaps recovery should go in world-load.
#
# This is implemented in the boot procedures found in the mud module code.

def StartRecovery():
    'Call this basically once per boot (from world load).'

    key = getenv(COPYOVERKEY)
    if key:
        global _active_copyover
        _active_copyover = Copyover(STUPHBIN, key)

        if _active_copyover.getStore().get('world'):
            # As last component:
            from stuphos.runtime import Component, eventResult
            class Recovery(Component):
                def onResetStart(self, ctlr):
                    # todo: uninstall this component:
                    return eventResult(True)

def CompleteRecovery():
    active = globals().get('_active_copyover')
    if active is not None:
        def recover():
            try: active.recover()
            except:
                syslog('Reboot Recovery N/A')
                from traceback import print_exc; print_exc()
            finally:
                del globals()['_active_copyover']

        from stuphos import enqueueHeartbeatTask
        enqueueHeartbeatTask(recover)

def Reboot(options):
    Copyover(STUPHBIN, DFLTKEY).initiate(options)


## -- Misc. --

def doReboot(*args):
    execl(STUPHBIN, EXENAME, '-server', *args)

## Spawn a new process that waits for its parent (this process) to die,
## whereby it executes another instance of stuph, possibly after a delay.
#
# This isn't necessary with supervisory program + auto-cycle.

# Command Overlay.
MINIMUM_COPYOVER_LEVEL = 115
def doCopyover(peer, cmd, argstr):
    args = argstr.split() if argstr else []
    if not args or args[0].lower() not in ('copy-over', 'copyover'):
        return False

    if peer.avatar and peer.avatar.level >= MINIMUM_COPYOVER_LEVEL:
        del args[0] # Not needed in freeze subsystem.

        from stuphmud.server.zones.freeze import getValidCmdlnOptions
        try: options = getValidCmdlnOptions(args)
        except SystemExit:
            return True

        from world.player import players
        for player in players():
            print('&r*&yInitiating Copyover, %s Saved&r*&N' % \
                  (player.avatar and player.avatar.name or 'Character'), file=player)

            print('Standby...', file=player)

        # XXX Not timed or confirmed!  (todo: modify act.wizard.cpp shutdown apparatus)
        from stuphos import enqueueHeartbeatTask
        enqueueHeartbeatTask(Reboot, options)
        return True

try: from stuphmud.server.player import ACMD
except ImportError: pass
else: ACMD('shutdown*')(doCopyover)



# Provides an import shelter for loading from certain package modules.
try: from game import lookAtRoom
except ImportError:
	# Remove reliance on game and instead rely on world.
	from world.player.commands import lookAtRoom



# Define verb commands (and others) based on mime message.
# Note: this probably makes more sense as a wm structure, but it looks more formal this way.
from email import message_from_string, message_from_file

STUPH_CODE_TEXT = 'text/stuph-code'

class MessageDecoder:
    def __init__(self, msg):
        self.msg = msg

    def __getitem__(self, name):
        return self.msg[name]
    def getPayload(self):
        return self.msg.get_payload()

    def getAllOptions(self, key):
        return self.msg.get_all(key)
    def get(self, key, default = None):
        return self.msg.get(key, default)

class StuphCode(MessageDecoder):
    def evaluate(self, **values):
        languages = self['X-Coding-Language'].split(';')
        languages = dict((n.strip().lower(), True) for n in languages)

        def isLanguage(name):
            try: return languages.pop(name)
            except KeyError:
                return False

        if isLanguage('player-command'):
            minlevel = self.get('X-Command-Level')
            group = self.get('X-Command-Group')
            verb = self.get('X-Command-Verb')

            # Currently only supported language.
            assert len(languages) is 1
            assert 'python' in languages
            language = 'python'

            options = self.parseCommandOptions(self.getAllOptions('X-Command-Options'))
            options = list(options)

            return self.definePlayerCommand(verb, language, self.getPayload(),
                                            options = options,
                                            minlevel = minlevel,
                                            group = group)

        raise NameError(', '.join(iter(languages.keys())))

    def parseCommandOptions(self, options):
        # <short> | <long>
        # <option> =
        #   true|false [, <dest>]
        #   store [, <dest>]
        #
        # I'm beginning to think Todd Downey was right.
        from stuphmud.server.player.commands import Option
        from stuphos.etc.tools import splitOne

        for opt in options or []:
            (opt, defn) = splitOne(opt, '=')
            if defn:
                opt = opt.strip()
                defn = [n.strip() for n in defn.split(',')]

                short = len(opt) == 1
                opt = ('-' + opt) if short else '--%s' % opt

                assert defn
                ln = defn[0].lower()

                # Option parameters.
                if ln in ['true', 'false']:
                    action = 'store_' + ln
                    if not short and len(defn) == 1:
                        yield Option(opt, action = action)

                    assert len(defn) == 2
                    yield Option(opt, action = action, dest = defn[1])

                elif ln == 'store':
                    assert len(defn) == 2
                    yield Option(opt, dest = defn[1])

                else:
                    raise SyntaxError('Unknown definition for option %s: %s' % (opt, defn))

            else:
                # Simple storage option.
                yield Option(opt)

    def definePlayerCommand(self, verb, language, source, **kwd):
        # Invoke lambda-programming api.
        from stuphmud.server.player.commands import programVerbCommand
        return programVerbCommand(verb, language, source, **kwd)

    class CommandBuilder:
        # Method chaining message builder.
        def __init__(self, verb = None, source = None, language = None):
            from email.message import Message
            self.msg = Message()
            self.msg.set_type(STUPH_CODE_TEXT)

            self.setVerb(verb)
            self.setSource(source)
            self.setLanguage(language)

        def __setitem__(self, name, value):
            self.msg[name] = value
        def setOnly(self, name, value):
            del self.msg[name]
            self[name] = value

        def setSource(self, source):
            if isinstance(source, (list, tuple)):
                source = '\n'.join(source)
            elif not source:
                source = ''

            self.msg.set_payload(source)
            return self

        def setVerb(self, verb):
            if verb:
                self.setOnly('X-Command-Verb', verb)

            return self

        def setLanguage(self, language):
            if language is None:
                language = 'python'
            else:
                assert language == 'python'

            self.setOnly('X-Coding-Language', 'player-command; ' + language)
            return self

        def setLevel(self, level):
            self.setOnly('X-Command-Level', level)
            return self

        def setGroup(self, group):
            self.setOnly('X-Command-Group', group)
            return self

        source = property(fset = setSource)
        verb = property(fset = setVerb)
        language = property(fset = setLanguage)
        level = property(fset = setLevel)
        group = property(fset = setGroup)

        def addOption(self, *args, **kwd):
            # Accepts regular add_option invocations and turns into X-Command-Option syntax.
            def _():
                for opt in args:
                    assert opt
                    if opt[0] == '-':
                        if len(opt) == 2:
                            opt = opt[1]
                        else:
                            assert len(opt) > 2 and opt[1] == '-'
                            opt = opt[2:]

                    action = kwd.get('action')
                    if action == ['store_true', 'store_false']:
                        opt = '%s=%s' % (opt, action[6:])

                        dest = kwd.get('dest')
                        if dest:
                            opt = '%s,%s' % (opt, dest)

                        yield opt

                    elif action:
                        dest = kwd.get('dest')
                        if dest:
                            yield '%s,%s' % (opt, dest)

                        yield opt

            for opt in _():
                self['X-Command-Option'] = opt

            return self

        # Message generation:
        def asString(self):
            return self.msg.as_string()
        def __str__(self):
            return self.asString()


# Todo: Runtime API.
MESSAGE_TYPE_EVALUTION = {STUPH_CODE_TEXT: StuphCode}
def loadMessage(msg, **kwd):
    V = MESSAGE_TYPE_EVALUTION[msg.get_content_type()]
    return V(msg).evaluate(**kwd)

def loadMessageString(string, **kwd):
    return loadMessage(message_from_string(string), **kwd)
def loadMessageFile(filename, **kwd):
    return loadMessage(message_from_file(filename), **kwd)


def testExample():
    '''
    MIME-Version: 1.0
    Content-Type: text/stuph-code
    X-Coding-Language: python; player-command
    X-Command-Level: implementor
    X-Command-Group: questing
    X-Command-Verb: do-this-*action
    X-Command-Option: v=true,verbose
    X-Command-Option: s=false,verbose
    X-Command-Option: verbose=true
    X-Command-Option: silent=false
    X-Command-Option: file-inputs
    X-Command-Option: f=store,file-inputs

    print '%r: %s' % (player, command)
    return True
    '''

    return StuphCode.CommandBuilder()                                          \
                    .setVerb('do-this-*action')                                \
                    .setLanguage('python')                                     \
                    .setLevel('implementor')                                   \
                    .setGroup('questing')                                      \
                    .addOption('-v', action = 'store_true', dest = 'verbose')  \
                    .addOption('-s', action = 'store_false', dest = 'verbose') \
                    .addOption('--verbose', action = 'store_true')             \
                    .addOption('--silent', action = 'store_false')             \
                    .addOption('--file-inputs')                                \
                    .addOption('-f', dest = 'file-inputs')                     \
                    .setSource(["print '%r: %s' % (player, command)",
                                "return True"])

# loadMessageString(str(testExample()))



#!/usr/bin/env python
from django.core.management import execute_manager
import imp
try:
    # Also, manually must construct registry because fraun.services.web/subdaemon import relies on it
    from stuphos.runtime.registry import getRegistry
    getRegistry(create = True)

    # Um, find_module apparently is more strict.
    # import pdb; pdb.set_trace()
    # imp.find_module('settings') # Assumed to be in the same directory.
except ImportError:
    import sys
    sys.stderr.write("Error: Can't find the file 'settings.py' in the directory containing %r. It appears you've customized things.\nYou'll have to run django-admin.py, passing it your settings module.\n(If the file settings.py does indeed exist, it's causing an ImportError somehow.)\n" % __file__)
    sys.exit(1)

from . import settings

if __name__ == "__main__":
    execute_manager(settings)

    # #!/usr/bin/env python
    # import os
    # import sys

    # if __name__ == "__main__":
    #     os.environ.setdefault("DJANGO_SETTINGS_MODULE", "b.settings")
    #     try:
    #         from django.core.management import execute_from_command_line
    #     except ImportError:
    #         # The above import may fail for some other reason. Ensure that the
    #         # issue is really that Django is missing to avoid masking other
    #         # exceptions on Python 2.
    #         try:
    #             import django
    #         except ImportError:
    #             raise ImportError(
    #                 "Couldn't import Django. Are you sure it's installed and "
    #                 "available on your PYTHONPATH environment variable? Did you "
    #                 "forget to activate a virtual environment?"
    #             )
    #         raise
    #     execute_from_command_line(sys.argv)
